import pytest

from bowtie._connectables import Connectable, UnknownConnector
from bowtie._containers import (
    IMAGE_REPOSITORY,
    ConnectableContainer,
    ConnectableImage,
)
from bowtie._direct_connectable import Direct
from bowtie.exceptions import CannotConnect

validators = Direct.from_id("python-jsonschema").registry()
validator = validators.for_uri("tag:bowtie.report,2024:connectables")
validated, invalidated = validator.validated, validator.invalidated


class TestImplicit:
    def test_with_repository(self):
        id = validated("example.com/bar")
        assert Connectable.from_str(id) == Connectable(
            id=id,
            connector=ConnectableImage(id="example.com/bar"),
        )

    def test_no_repository(self):
        id = validated("bar")
        assert Connectable.from_str(id) == Connectable(
            id=id,
            connector=ConnectableImage(id=f"{IMAGE_REPOSITORY}/bar"),
        )

    def test_with_repository_and_tag(self):
        id = validated("example.com/bar:latest")
        assert Connectable.from_str(id) == Connectable(
            id=id,
            connector=ConnectableImage(id="example.com/bar:latest"),
        )

    def test_known_direct(self):
        name = validated("python-jsonschema")
        direct = validated(f"direct:{name}")
        assert Connectable.from_str(name) == Connectable.from_str(direct)

    def test_unknown_direct(self):
        name = validated("asdf")
        image = validated(f"image:{name}")
        assert Connectable.from_str(name) == Connectable.from_str(image)


class TestImage:
    def test_with_repository(self):
        id = validated("image:foo/bar")
        assert Connectable.from_str(id) == Connectable(
            id=id,
            connector=ConnectableImage(id="foo/bar"),
        )

    def test_with_repository_and_tag(self):
        id = validated("image:foo/bar:latest")
        assert Connectable.from_str(id) == Connectable(
            id=id,
            connector=ConnectableImage(id="foo/bar:latest"),
        )

    def test_no_repository(self):
        id = validated("image:bar")
        assert Connectable.from_str(id) == Connectable(
            id=id,
            connector=ConnectableImage(id=f"{IMAGE_REPOSITORY}/bar"),
        )

    def test_read_timeout_sec(self):
        id = validated("image:bar:read_timeout_sec=5")
        assert Connectable.from_str(id) == Connectable(
            id=id,
            connector=ConnectableImage(
                id=f"{IMAGE_REPOSITORY}/bar",
                read_timeout_sec=5,
            ),
        )

    def test_no_timeout(self):
        id = validated("image:bar:read_timeout_sec=0")
        assert Connectable.from_str(id) == Connectable(
            id=id,
            connector=ConnectableImage(
                id=f"{IMAGE_REPOSITORY}/bar",
                read_timeout_sec=None,
            ),
        )


class TestContainer:
    def test_uuid(self):
        id = validated("container:c7895a98f49d")
        assert Connectable.from_str(id) == Connectable(
            id=id,
            connector=ConnectableContainer(id="c7895a98f49d"),
        )

    def test_read_timeout_sec(self):
        id = validated("container:c7895a98f49d:read_timeout_sec=5")
        assert Connectable.from_str(id) == Connectable(
            id=id,
            connector=ConnectableContainer(
                id="c7895a98f49d",
                read_timeout_sec=5,
            ),
        )

    def test_no_timeout(self):
        id = validated("container:c7895a98f49d:read_timeout_sec=0")
        assert Connectable.from_str(id) == Connectable(
            id=id,
            connector=ConnectableContainer(
                id="c7895a98f49d",
                read_timeout_sec=None,
            ),
        )


class TestDirect:
    def test_named_python_jsonschema(self):
        id = validated("direct:python-jsonschema")
        assert Connectable.from_str(id) == Connectable(
            id=id,
            connector=Direct.from_id(id="python-jsonschema"),
        )

    def test_null(self):
        id = validated("direct:null")
        assert Connectable.from_str(id) == Connectable(
            id=id,
            connector=Direct.null(),
        )

    def test_null_considers_everything_valid(self):
        assert Direct.null().registry().for_schema(False).is_valid(37)

    def test_named_unknown(self):
        id = validated("direct:foobar")
        with pytest.raises(CannotConnect, match="'foobar'"):
            Connectable.from_str(id)

    def test_import(self):
        id = validated("direct:bowtie._direct_connectable:jsonschema")

        from bowtie._direct_connectable import jsonschema

        assert Connectable.from_str(id) == Connectable(
            id=id,
            connector=Direct(wraps=jsonschema()),
        )

    def test_import_parameters(self):
        id = validated(
            "direct:bowtie.tests.miniatures:only_supports,dialect=draft7",
        )

        from bowtie.tests.miniatures import only_supports

        assert Connectable.from_str(id) == Connectable(
            id=id,
            connector=Direct(
                wraps=only_supports(dialect="draft7"),
            ),
        )

    def test_import_missing_parameter(self):
        """
        The basic import syntax is direct:foo.bar:baz, not foo.bar.baz.
        """
        prefix, suffix = "direct:bowtie.tests.miniatures", "always_invalid"

        Connectable.from_str(validated(f"{prefix}:{suffix}"))  # succeeds

        with pytest.raises(CannotConnect, match=":always_invalid'"):
            Connectable.from_str(invalidated(f"{prefix}.{suffix}"))


class TestExplicitHappy:
    def test_known_direct(self):
        name = "python-jsonschema"
        happy = validated(f"happy:{name}")
        direct = validated(f"direct:{name}")
        assert Connectable.from_str(happy) == Connectable.from_str(direct)

    def test_unknown_direct_becomes_image(self):
        name = "asdf"
        happy = validated(f"happy:{name}")
        image = validated(f"image:{name}")
        assert Connectable.from_str(happy) == Connectable.from_str(image)

    def test_image_parameters(self):
        name = "asdf:read_timeout_sec=5"
        happy = validated(f"happy:{name}")
        image = validated(f"image:{name}")
        assert Connectable.from_str(happy) == Connectable.from_str(image)


def test_unknown_connector():
    id = invalidated("asdf:foo")
    with pytest.raises(UnknownConnector):
        Connectable.from_str(id)


class TestToTerse:
    def test_explicit_image_with_repository(self):
        id = validated("image:foo/bar")
        assert Connectable.from_str(id).to_terse() == "foo/bar"

    def test_explicit_image_bowtie_repository(self):
        id = validated("image:ghcr.io/bowtie-json-schema/bar")
        assert Connectable.from_str(id).to_terse() == "bar"

    def test_explicit_image_no_repository(self):
        id = validated("image:bar")
        assert Connectable.from_str(id).to_terse() == "bar"

    def test_implicit_image_with_repository(self):
        id = validated("foo/bar")
        assert Connectable.from_str(id).to_terse() == "foo/bar"

    def test_implicit_image_bowtie_repository(self):
        id = validated("ghcr.io/bowtie-json-schema/bar")
        assert Connectable.from_str(id).to_terse() == "bar"

    def test_implicit_image_no_repository(self):
        id = validated("bar")
        assert Connectable.from_str(id).to_terse() == "bar"



from pathlib import Path
import uuid

from url.url import URL

from bowtie._benchmarks import Benchmark, BenchmarkGroup
from bowtie._core import Dialect


def get_benchmark():
    name = "required"
    benchmark_type = "keyword"
    description = (
        "A benchmark for measuring performance of the "
        "implementation for the required keyword."
    )
    max_array_length = 100000
    varying_parameter = "Array length"

    array_length = 1000
    benchmarks = []
    while array_length * 2 <= max_array_length:
        num_required_properties = array_length // 2
        required_properties = [
            uuid.uuid4().hex for _ in range(num_required_properties)
        ]

        object_keys = [uuid.uuid4().hex for _ in range(array_length)]
        all_at_beginning = required_properties + object_keys
        all_at_middle = (
            object_keys[: array_length // 2]
            + required_properties
            + object_keys[array_length // 2 :]
        )
        all_at_last = object_keys + required_properties
        none_present = object_keys + object_keys

        tests = (
            [
                dict(
                    description="All required properties at beginning",
                    instance=_generate_object_with_keys(all_at_beginning),
                ),
                dict(
                    description="All required properties at middle",
                    instance=_generate_object_with_keys(all_at_middle),
                ),
                dict(
                    description="All required properties at last",
                    instance=_generate_object_with_keys(all_at_last),
                ),
                dict(
                    description="None of the required properties present",
                    instance=_generate_object_with_keys(none_present),
                ),
            ]
            if array_length == max_array_length
            else [
                dict(
                    description="None of the required properties present",
                    instance=_generate_object_with_keys(none_present),
                ),
            ]
        )

        benchmarks.append(
            Benchmark.from_dict(
                name=f"Array length - {array_length}",
                description=(
                    f"Validating the `required` keyword over array of length {array_length}."
                ),
                schema={
                    "type": "object",
                    "required": required_properties,
                },
                tests=tests,
            ),
        )
        array_length *= 10

    return BenchmarkGroup(
        name=name,
        benchmark_type=benchmark_type,
        description=description,
        dialects_supported=[
            Dialect.from_str("https://json-schema.org/draft/2020-12/schema"),
            Dialect.from_str("https://json-schema.org/draft/2019-09/schema"),
            Dialect.from_str("http://json-schema.org/draft-07/schema#"),
            Dialect.from_str("http://json-schema.org/draft-06/schema#"),
            Dialect.from_str("http://json-schema.org/draft-04/schema#"),
            Dialect.from_str("http://json-schema.org/draft-03/schema#"),
        ],
        benchmarks=benchmarks,
        uri=URL.parse(Path(__file__).absolute().as_uri()),
        varying_parameter=varying_parameter,
    )


def _generate_object_with_keys(keys):
    return {key: uuid.uuid4().hex for key in keys}



from pathlib import Path
import uuid

from url.url import URL

from bowtie._benchmarks import Benchmark, BenchmarkGroup
from bowtie._core import Dialect


def get_benchmark():
    name = "items"
    benchmark_type = "keyword"
    description = (
        "A benchmark for measuring performance of the "
        "implementation for the items keyword."
    )
    max_array_length = 100000
    varying_parameter = "Array length"

    array_length = 1000
    benchmarks = []
    while array_length <= max_array_length:
        array = [uuid.uuid4().hex for _ in range(array_length)]

        invalid_at_first = [1] + array[:-1]
        invalid_at_middle = (
            array[: array_length // 2] + [1] + array[array_length // 2 : -1]
        )
        invalid_at_last = array[:-1] + [1]
        valid = array

        tests = (
            [
                dict(
                    description="Invalid at First",
                    instance=invalid_at_first,
                ),
                dict(
                    description="Invalid at Middle",
                    instance=invalid_at_middle,
                ),
                dict(
                    description="Invalid at Last",
                    instance=invalid_at_last,
                ),
                dict(description="Valid", instance=valid),
            ]
            if array_length == max_array_length
            else [
                dict(description="Valid", instance=valid),
            ]
        )

        benchmarks.append(
            Benchmark.from_dict(
                name=f"Array length - {array_length}",
                description=(
                    f"Validating the `items` keyword over array of length {array_length}."
                ),
                schema={
                    "type": "array",
                    "items": {"type": "string"},
                },
                tests=tests,
            ),
        )
        array_length *= 10

    return BenchmarkGroup(
        name=name,
        benchmark_type=benchmark_type,
        description=description,
        dialects_supported=[
            Dialect.from_str("https://json-schema.org/draft/2020-12/schema"),
            Dialect.from_str("https://json-schema.org/draft/2019-09/schema"),
            Dialect.from_str("http://json-schema.org/draft-07/schema#"),
            Dialect.from_str("http://json-schema.org/draft-06/schema#"),
            Dialect.from_str("http://json-schema.org/draft-04/schema#"),
            Dialect.from_str("http://json-schema.org/draft-03/schema#"),
        ],
        benchmarks=benchmarks,
        uri=URL.parse(Path(__file__).absolute().as_uri()),
        varying_parameter=varying_parameter,
    )



from pathlib import Path
import random
import string
import uuid

from url.url import URL

from bowtie._benchmarks import Benchmark, BenchmarkGroup
from bowtie._core import Dialect


def get_benchmark():
    name = "propertyNames"
    benchmark_type = "keyword"
    description = (
        "A benchmark for measuring performance of the "
        "implementation for the propertyNames keyword."
    )
    max_object_length = 100000
    varying_parameter = "Object length"

    object_length = 1000
    benchmarks = []
    while object_length <= max_object_length:
        object = {uuid.uuid4().hex: 10 for _ in range(object_length)}
        invalid_property = "".join(
            random.choice(string.ascii_letters) for _ in range(7)
        )

        invalid_at_first = {invalid_property: 10}
        invalid_at_first.update(object)

        invalid_at_middle = {
            uuid.uuid4().hex: 10 for _ in range(object_length // 2)
        }
        invalid_at_middle.update({invalid_property: 10})
        invalid_at_middle.update(
            {uuid.uuid4().hex: 10 for _ in range(object_length // 2)},
        )

        invalid_at_last = object.copy()
        invalid_at_last.update({invalid_property: 10})

        valid = {
            "".join(random.choice(string.ascii_letters) for _ in range(1)): 10
            for _ in range(object_length)
        }

        tests = (
            [
                dict(
                    description="Invalid at First",
                    instance=invalid_at_first,
                ),
                dict(
                    description="Invalid at Middle",
                    instance=invalid_at_middle,
                ),
                dict(
                    description="Invalid at Last",
                    instance=invalid_at_last,
                ),
                dict(description="Valid", instance=valid),
            ]
            if object_length == max_object_length
            else [
                dict(description="Valid", instance=valid),
            ]
        )

        benchmarks.append(
            Benchmark.from_dict(
                name=f"Num of Properties - {object_length}",
                description=(
                    f"Validating the `propertyNames` keyword over object of length {object_length}."
                ),
                schema={"propertyNames": {"maxLength": 5}},
                tests=tests,
            ),
        )
        object_length *= 10

    return BenchmarkGroup(
        name=name,
        benchmark_type=benchmark_type,
        description=description,
        dialects_supported=[
            Dialect.from_str("https://json-schema.org/draft/2020-12/schema"),
            Dialect.from_str("https://json-schema.org/draft/2019-09/schema"),
            Dialect.from_str("http://json-schema.org/draft-07/schema#"),
            Dialect.from_str("http://json-schema.org/draft-06/schema#"),
        ],
        benchmarks=benchmarks,
        uri=URL.parse(Path(__file__).absolute().as_uri()),
        varying_parameter=varying_parameter,
    )



from datetime import date

import pytest

from bowtie._core import Dialect


def test_known_latest():
    assert Dialect.latest() in Dialect.known()


def test_latest():
    assert Dialect.latest().pretty_name == "Draft 2020-12"


def test_top():
    validator = Dialect.latest().top()
    assert validator.is_valid(37)


def test_bottom():
    validator = Dialect.latest().bottom()
    assert not validator.is_valid(37)


def test_top_draft3():
    validator = Dialect.by_short_name()["draft3"].top()
    assert validator.is_valid(37)


def test_bottom_draft3():
    validator = Dialect.by_short_name()["draft3"].bottom()
    assert not validator.is_valid(37)


def test_no_top():
    no_top = Dialect(
        pretty_name="No top!",
        short_name="NoTop",
        uri="urn:example:no-top",
        first_publication_date=date.today(),
        top_schema=None,
    )
    with pytest.raises(ValueError, match="has no top"):
        no_top.top()


def test_no_bottom():
    no_bottom = Dialect(
        pretty_name="No bottom!",
        short_name="NoBottom",
        uri="urn:example:no-bottom",
        first_publication_date=date.today(),
        bottom_schema=None,
    )
    with pytest.raises(ValueError, match="has no bottom"):
        no_bottom.bottom()



from __future__ import annotations

from datetime import UTC, datetime
from pathlib import Path
from statistics import geometric_mean
from typing import TYPE_CHECKING
import asyncio
import importlib
import importlib.metadata
import json
import os
import statistics
import subprocess
import tempfile

from attrs import asdict, field, frozen
from attrs.filters import exclude
from diagnostic import DiagnosticError
from rich import box
from rich.console import Console
from rich.progress import Progress
from rich.table import Column, Table
from url import URL
import pyperf  # type: ignore[reportMissingTypeStubs]

from bowtie import _report
from bowtie._core import (
    Dialect,
    Example,
    ImplementationInfo,
    Test,
    TestCase,
    convert_table_to_markdown,
)
from bowtie._direct_connectable import Direct

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping, Sequence
    from typing import Any

    from bowtie._commands import (
        Message,
    )
    from bowtie._connectables import Connectable, ConnectableId

Seq = int | str

Benchmark_Group_Name = str
Benchmark_Criteria = str

STDOUT = Console()
STDERR = Console(stderr=True)

benchmark_validator = (
    Direct.from_id("python-jsonschema")
    .registry()
    .for_uri(
        URL.parse("tag:bowtie.report,2024:benchmarks"),
    )
)
benchmark_validated, benchmark_invalidated = (
    benchmark_validator.validated,
    benchmark_validator.invalidated,
)


def get_benchmark_files(
    benchmark_type: str | None,
    dialect: Dialect,
    benchmarks: Iterable[str] = [],
) -> list[Path]:
    bowtie_parent_dir = Path(__file__).parent.parent
    benchmarks_dir = bowtie_parent_dir / "bowtie/benchmarks"
    files: list[Path] = []

    for file in benchmarks_dir.rglob("*"):
        relative_folder_path = file.parent.relative_to(bowtie_parent_dir)
        module_name = str(relative_folder_path).replace(os.sep, ".")
        benchmark_group = _load_benchmark_group_from_file(
            file,
            module_name,
        )
        if (
            benchmark_group is not None
            and (
                benchmark_group.benchmark_type == benchmark_type
                or benchmark_type is None
            )
            and dialect in benchmark_group.dialects_supported
        ):
            files.append(file)

    if benchmarks:
        files = [
            matched_file
            for benchmark in benchmarks
            if (
                matched_file := next(
                    (file for file in files if benchmark in str(file)),
                    None,
                )
            )
        ]

    return files


def combine_benchmark_reports_to_serialized(
    benchmark_report_files: Iterable[Path],
) -> str:
    merged_report = BenchmarkReport.merge(
        [
            BenchmarkReport.from_dict(
                **json.loads(report_file.read_text()),
            )
            for report_file in benchmark_report_files
        ],
    )
    return json.dumps(merged_report.serializable())


def _load_benchmark_group_from_file(
    file: Path,
    module: str | None = None,
):
    benchmark_group = None
    if file.suffix == ".py" and file.name != "__init__.py":
        benchmark_module_name = "." + file.stem
        loaded_class = importlib.import_module(
            benchmark_module_name,
            module,
        ).get_benchmark()
        if isinstance(loaded_class, Benchmark):
            benchmark_group = BenchmarkGroup.from_dict(
                loaded_class.serializable(),
                uri=URL.parse(file.absolute().as_uri()),
            )
        elif isinstance(loaded_class, BenchmarkGroup):
            benchmark_group = loaded_class

    elif file.suffix == ".json":
        data = json.loads(file.read_text())
        benchmark_group = BenchmarkGroup.from_dict(
            data,
            uri=URL.parse(file.absolute().as_uri()),
        )

    return benchmark_group


@frozen
class BenchmarkLoadError(Exception):
    message: str

    def __rich__(self):
        return DiagnosticError(
            code="benchmark-load-error",
            message=self.message,
            causes=[],
            hint_stmt=(
                "Make sure that the benchmarks are present "
                "in the appropriate folder and follow the specified schema."
            ),
        )


class BenchmarkError(Exception):
    pass


@frozen
class PyperfError(BenchmarkError):
    """
    Some Error with Pyperf while running the Benchmark.
    """

    error_stack: str

    def __rich__(self):
        return DiagnosticError(
            code="pyperf-benchmark-error",
            message="Pyperf command failed while running the Benchmark.",
            note_stmt=self.error_stack,
            causes=[],
            hint_stmt=(
                "Make sure that pyperf is installed and "
                "all the arguments are provided correctly."
            ),
        )


@frozen
class BowtieRunError(BenchmarkError):
    """
    Some Error with Bowtie while running the Benchmark.
    """

    error_stack: str
    connectable_id: ConnectableId | None

    def __rich__(self):
        return DiagnosticError(
            code="bowtie-benchmark-error",
            message="Bowtie Run command failed while running the Benchmark.",
            note_stmt=self.error_stack,
            causes=[],
            hint_stmt=(
                "Some internal error has occured while running the Benchmark "
                f"with {self.connectable_id}. (Probably some error with "
                "the harness)"
            ),
        )


@frozen
class Benchmark:
    description: str
    name: str
    schema: Any
    tests: Sequence[Example | Test]
    dialect: Dialect | None = None

    @classmethod
    def from_dict(
        cls,
        tests: Iterable[dict[str, Any]],
        name: str,
        **kwargs: Any,
    ):
        return cls(
            tests=[Example.from_dict(**test) for test in tests],
            name=name,
            **kwargs,
        )

    def serializable(self) -> Message:
        return asdict(
            self,
            filter=lambda _, v: v is not None,
        )

    def benchmark_with_diff_tests(self, tests: Sequence[Example | Test]):
        benchmark = self.serializable()
        benchmark["tests"] = tests
        return Benchmark(**benchmark)

    def maybe_set_dialect_from_schema(self):
        dialect_from_schema: str | None = (  # type: ignore[reportUnknownVariableType]
            self.schema.get("$schema")  # type: ignore[reportUnknownMemberType]
            if isinstance(self.schema, dict)
            else None
        )
        if not dialect_from_schema:
            return self

        benchmark = self.serializable()
        benchmark["dialect"] = Dialect.from_str(
            dialect_from_schema,  # type: ignore[reportUnknownArgumentType]
        )
        return Benchmark.from_dict(**benchmark)


@frozen
class BenchmarkGroup:
    name: str
    benchmarks: Sequence[Benchmark]
    description: str
    uri: URL | None
    benchmark_type: str
    dialects_supported: Sequence[Dialect] = list(Dialect.known())
    varying_parameter: str | None = None

    @classmethod
    def from_folder(
        cls,
        folder: Path,
        module: str = "bowtie.benchmarks",
    ) -> Iterable[BenchmarkGroup]:
        for file in folder.iterdir():
            benchmark_group = cls.from_file(file, module)
            if not benchmark_group:
                continue
            yield benchmark_group

    @classmethod
    def from_file(
        cls,
        file: Path,
        module: str = "bowtie.benchmarks",
    ) -> BenchmarkGroup | None:
        return _load_benchmark_group_from_file(
            file,
            module,
        )

    @classmethod
    def from_dict(
        cls,
        data: dict[str, Any],
        uri: URL | None = None,
    ) -> BenchmarkGroup:
        varying_parameter = data.pop("varying_parameter", None)
        benchmark_type = data.pop("benchmark_type", "None")
        dialects_supported = [
            (
                dialect
                if isinstance(dialect, Dialect)
                else Dialect.from_str(dialect)
            )
            for dialect in data.pop("dialects_supported", Dialect.known())
        ]

        if "benchmarks" not in data:
            benchmark = Benchmark.from_dict(
                **data,
            ).maybe_set_dialect_from_schema()
            return cls(
                name=benchmark.name,
                description=benchmark.description,
                benchmarks=[benchmark],
                uri=uri,
                benchmark_type=benchmark_type,
                dialects_supported=dialects_supported,
                varying_parameter=varying_parameter,
            )

        benchmarks = [
            Benchmark.from_dict(
                **benchmark,
            ).maybe_set_dialect_from_schema()
            for benchmark in data["benchmarks"]
        ]
        return cls(
            name=data["name"],
            description=data["description"],
            benchmarks=benchmarks,
            uri=uri,
            benchmark_type=benchmark_type,
            varying_parameter=varying_parameter,
            dialects_supported=dialects_supported,
        )

    def serializable(self) -> Message:
        serialized_dict = asdict(
            self,
            filter=lambda _, v: v is not None,
        )
        if "uri" in serialized_dict:
            serialized_dict["uri"] = str(serialized_dict["uri"])
        serialized_dict["dialects_supported"] = [
            dialect.serializable() for dialect in self.dialects_supported
        ]
        return serialized_dict


@frozen
class BenchmarkGroupResult:
    name: str
    benchmark_type: str
    description: str
    benchmark_results: list[BenchmarkResult]
    varying_parameter: str | None = None

    def serializable(self):
        return asdict(self)

    @classmethod
    def from_dict(
        cls,
        benchmark_results: list[dict[str, Any]],
        **kwargs: Any,
    ) -> BenchmarkGroupResult:
        return cls(
            benchmark_results=[
                BenchmarkResult.from_dict(**result)
                for result in benchmark_results
            ],
            **kwargs,
        )


@frozen
class BenchmarkResult:
    name: str
    description: str
    test_results: Sequence[TestResult]

    def serializable(self):
        return asdict(self)

    @classmethod
    def from_dict(
        cls,
        test_results: list[dict[str, Any]],
        **kwargs: Any,
    ) -> BenchmarkResult:
        return cls(
            test_results=[
                TestResult.from_dict(**result) for result in test_results
            ],
            **kwargs,
        )


@frozen
class TestResult:
    description: str
    connectable_results: Sequence[ConnectableResult]

    def serializable(self):
        return asdict(self)

    @classmethod
    def from_dict(
        cls,
        connectable_results: list[dict[str, Any]],
        **kwargs: Any,
    ) -> TestResult:
        return cls(
            connectable_results=[
                ConnectableResult.from_dict(**result)
                for result in connectable_results
            ],
            **kwargs,
        )


@frozen
class ConnectableResult:
    connectable_id: ConnectableId
    duration: float
    values: list[float]
    errored: bool = False

    def serializable(self):
        return asdict(self, filter=lambda _, v: v is not None)

    @classmethod
    def from_dict(cls, **kwargs: Any) -> ConnectableResult:
        return cls(
            **kwargs,
        )


@frozen
class BenchmarkMetadata:
    dialect: Dialect
    implementations: Mapping[ConnectableId, ImplementationInfo] = field(
        alias="implementations",
    )
    num_runs: int
    num_values: int
    num_warmups: int
    num_loops: int
    system_metadata: dict[str, Any] = field(factory=dict, repr=False)

    bowtie_version: str = field(
        default=importlib.metadata.version("bowtie-json-schema"),
        eq=False,
        repr=False,
    )
    started: datetime = field(
        factory=lambda: datetime.now(UTC),
        eq=False,
        repr=False,
    )

    def serializable(self):
        as_dict = asdict(
            self,
            filter=exclude("dialect"),
            recurse=False,
        )
        as_dict.update(
            dialect=self.dialect.serializable(),
            started=as_dict.pop("started").isoformat(),
            implementations={
                id: implementation.serializable()
                for id, implementation in self.implementations.items()
            },
        )
        return as_dict

    @classmethod
    def from_dict(
        cls,
        dialect: str,
        implementations: dict[str, dict[str, Any]],
        started: str | None = None,
        **kwargs: Any,
    ) -> BenchmarkMetadata:
        if started is not None:
            kwargs["started"] = datetime.fromisoformat(started)
        return cls(
            dialect=Dialect.from_str(dialect),
            implementations={
                id: ImplementationInfo.from_dict(**data)
                for id, data in implementations.items()
            },
            **kwargs,
        )


@frozen
class BenchmarkReport:
    metadata: BenchmarkMetadata
    results: dict[Benchmark_Group_Name, BenchmarkGroupResult] = field(
        factory=dict,
    )

    def serializable(self):
        as_dict: dict[str, Any] = dict(metadata=self.metadata.serializable())
        as_dict["results"] = [
            benchmark_group_result.serializable()
            for _, benchmark_group_result in self.results.items()
        ]
        return as_dict

    @classmethod
    def from_dict(
        cls,
        metadata: dict[str, Any],
        results: list[dict[str, Any]],
    ) -> BenchmarkReport:
        return cls(
            metadata=BenchmarkMetadata.from_dict(**metadata),
            results={
                r["name"]: BenchmarkGroupResult.from_dict(**r) for r in results
            },
        )

    @classmethod
    def merge(
        cls,
        benchmark_reports: list[BenchmarkReport],
    ) -> BenchmarkReport:
        return BenchmarkReport(
            metadata=benchmark_reports[0].metadata,
            results={
                k: v
                for report in benchmark_reports
                for k, v in report.results.items()
            },
        )


@frozen
class BenchmarkReporter:
    _progress_bar: Progress = field(alias="progress_bar")
    _report: BenchmarkReport = field(alias="report")

    _quiet: bool = field(alias="quiet", default=False)
    _format: str = field(alias="format", default="pretty")
    _benchmark_group_uri: dict[str, URL | None] = field(
        factory=dict,
    )
    _mean_threshold: float = field(alias="mean_threshold", default=0.10)

    def update_system_metadata(self, system_metadata: dict[str, Any]):
        system_metadata.pop("command")
        system_metadata.pop("name")
        self._report.metadata.system_metadata.update(system_metadata)

    def started(self):
        pass

    def report_incompatible_connectables(
        self,
        incompatible_connectables: Sequence[Connectable],
        dialect: Dialect,
    ):
        for connectable in incompatible_connectables:
            if not self._quiet:
                STDOUT.log(
                    f"{connectable.to_terse()} does not support "
                    f"dialect {dialect.serializable()}\n",
                )

    @staticmethod
    def _total_tests_in_benchmark_group(benchmark_group: BenchmarkGroup):
        num = 0
        for benchmark in benchmark_group.benchmarks:
            num += len(benchmark.tests)
        return num

    def running_benchmark_group(
        self,
        benchmark_group: BenchmarkGroup,
    ):
        progress_bar_task = None
        if not self._quiet:
            progress_bar_task = self._progress_bar.add_task(
                f"Running Benchmark Group: {benchmark_group.name}",
                total=len(self._report.metadata.implementations)
                * self._total_tests_in_benchmark_group(benchmark_group),
            )
        benchmark_results: list[BenchmarkResult] = []
        self._benchmark_group_uri[benchmark_group.name] = benchmark_group.uri

        def benchmark_started(benchmark_name: str, benchmark_description: str):
            test_results: list[TestResult] = []

            def test_started(test_description: str):
                connectable_results: dict[
                    ConnectableId,
                    ConnectableResult,
                ] = {}

                def got_connectable_result(
                    connectable: ConnectableId,
                    test_result: str,
                    retry_count: int,
                    measured_time_values: list[float],
                    errored: bool = False,
                ):
                    retry_needed = False

                    if errored:
                        if not self._quiet:
                            STDOUT.log(
                                f"WARNING!\n{benchmark_group.name}:"
                                f"{benchmark_name}:{test_description}\n"
                                f"Some Error was encountered while running"
                                f"test for {connectable}\n ",
                            )
                        connectable_result = ConnectableResult(
                            connectable_id=connectable,
                            duration=0,
                            values=[1e9, 1e9],
                            errored=True,
                        )
                        connectable_results[connectable] = connectable_result

                    else:
                        benchmark_result = pyperf.Benchmark.loads(test_result)  # type: ignore[reportUnknownVariableType]

                        # Ignoring Warmup Values
                        result_values = measured_time_values[
                            self._report.metadata.num_warmups
                            * self._report.metadata.num_runs :
                        ]

                        if not len(result_values):
                            raise BowtieRunError(
                                "No results were returned",
                                connectable,
                            )

                        benchmark_duration: float = float(
                            benchmark_result.get_total_duration(),  # type: ignore[reportUnknownMemberType]
                        )
                        connectable_result = ConnectableResult(
                            connectable_id=connectable,
                            duration=benchmark_duration,
                            values=result_values,
                        )

                        mean = statistics.mean(result_values)
                        std_dev = statistics.stdev(result_values)
                        if (
                            std_dev / mean > self._mean_threshold
                            and not self._quiet
                        ):
                            retry_needed = True
                            if retry_count == 0:
                                STDOUT.log(
                                    f"WARNING!\n{benchmark_name}:"
                                    f"{test_description}:{connectable}\n"
                                    f"Benchmark Might Be Unstable "
                                    f"(std_dev = {(std_dev / mean) * 100}%)",
                                )

                        if not len(self._report.metadata.system_metadata):
                            self.update_system_metadata(
                                benchmark_result.get_metadata(),  # type: ignore[reportUnknownMemberType]
                            )
                        connectable_results[connectable] = connectable_result

                    if progress_bar_task is not None:
                        if benchmark_group.name == benchmark_name:
                            self._progress_bar.update(
                                progress_bar_task,
                                description=(
                                    f"Running Benchmark: {benchmark_name}\n"
                                    f"Test: {test_description}\n"
                                    f"Connectable: {connectable}"
                                ),
                                advance=1,
                            )
                        else:
                            self._progress_bar.update(
                                progress_bar_task,
                                description=(
                                    f"Running Benchmark Group: "
                                    f"{benchmark_group.name}\n"
                                    f"Benchmark: {benchmark_name}\n"
                                    f"Test: {test_description}\n"
                                    f"Connectable: {connectable}"
                                ),
                                advance=1,
                            )

                    return retry_needed

                def test_finished():
                    test_result = TestResult(
                        description=test_description,
                        connectable_results=list(connectable_results.values()),
                    )
                    test_results.append(test_result)

                return got_connectable_result, test_finished

            def benchmark_finished():
                benchmark_result = BenchmarkResult(
                    name=benchmark_name,
                    description=benchmark_description,
                    test_results=test_results,
                )
                benchmark_results.append(benchmark_result)

            return test_started, benchmark_finished

        def benchmark_group_finished():
            benchmark_group_result = BenchmarkGroupResult(
                name=benchmark_group.name,
                benchmark_type=benchmark_group.benchmark_type,
                description=benchmark_group.description,
                benchmark_results=benchmark_results,
                varying_parameter=benchmark_group.varying_parameter,
            )
            self._report.results[benchmark_group.name] = benchmark_group_result
            if progress_bar_task is not None:
                self._progress_bar.update(progress_bar_task, visible=False)

        return benchmark_started, benchmark_group_finished

    def finished(self):
        if self._format in ("pretty", "markdown"):
            self._print_results_table()
        else:
            STDOUT.print_json(data=self._report.serializable())

    def no_compatible_connectables(self):
        if not self._quiet:
            STDOUT.log("Skipping Benchmark, No Connectables to run !")

    def _print_results_table(self):

        def _format_value(value: float) -> str:

            if value * 1000 < 1:
                return f"{round(value * 1000 * 1000)}us"
            elif value < 1:
                return f"{round(value * 1000)}ms"
            return f"{round(value, 2)}s"

        def _get_sorted_table_from_test_results(
            test_results: Sequence[TestResult],
        ):
            results_for_connectable: dict[
                ConnectableId,
                list[tuple[float, float, bool]],
            ] = {}
            ref_row: list[str] = [""]

            for idx, connectable_result in enumerate(
                test_results[0].connectable_results,
            ):
                connectable_results = [
                    (
                        statistics.mean(
                            test_result.connectable_results[idx].values,
                        ),
                        statistics.stdev(
                            test_result.connectable_results[idx].values,
                        ),
                        test_result.connectable_results[idx].errored,
                    )
                    for test_result in test_results
                ]
                results_for_connectable[connectable_result.connectable_id] = (
                    connectable_results
                )

            sorted_connectables = sorted(
                results_for_connectable.keys(),
                key=(
                    lambda connectable_id: (
                        geometric_mean(
                            [
                                result_mean
                                for (
                                    result_mean,
                                    _,
                                    errored,
                                ) in results_for_connectable[connectable_id]
                                if not errored
                            ]
                            or [1e9],
                        )
                    )
                ),
            )
            results_for_connectable = {
                connectable: results_for_connectable[connectable]
                for connectable in sorted_connectables
            }
            columns = ["Test Name"]
            rows: list[list[str]] = []
            for (
                connectable_id,
                connectable_results,
            ) in results_for_connectable.items():
                columns.append(
                    connectable_id,
                )
                g_mean = geometric_mean(
                    [
                        result_mean
                        for result_mean, _, errored in connectable_results
                        if not errored
                    ]
                    or [1e9],
                )
                if g_mean == geometric_mean([1e9]):
                    ref_row.append("Errored")
                else:
                    ref_row.append(
                        str(g_mean),
                    )

            fastest_connectable_results = next(
                iter(
                    results_for_connectable.values(),
                ),
            )

            for idx, test_result in enumerate(
                test_results,
            ):
                row_elements = [
                    test_result.description,
                ]
                for connectable_idx, results in enumerate(
                    results_for_connectable.items(),
                ):
                    _, connectable_results = results
                    _mean, std_dev, errored = connectable_results[idx]

                    fastest_implementation_mean, _, __ = (
                        fastest_connectable_results[idx]
                    )
                    relative = _mean / fastest_implementation_mean

                    if errored:
                        repr_string = "Errored"
                    else:
                        repr_string = (
                            f"{_format_value(_mean)} +- "
                            f"{_format_value(std_dev)}"
                        )
                        if connectable_idx != 0 and relative > 1:
                            repr_string += f": {round(relative, 2)}x slower"
                        elif connectable_idx:
                            repr_string += (
                                f": {round(1 / relative, 2)}x faster"
                            )

                    row_elements.append(repr_string)

                rows.append(row_elements)

            for implementation_idx in range(2, len(ref_row)):
                if ref_row[implementation_idx] == "Errored":
                    continue
                rounded_off_val = round(
                    float(ref_row[implementation_idx]) / float(ref_row[1]),
                    2,
                )
                ref_row[implementation_idx] = f"{rounded_off_val}x slower"

            ref_row[1] = "Reference"

            if len(results_for_connectable) > 1:
                rows.append(ref_row)

            return rows, columns

        def _table_for_benchmark_result(
            benchmark_result: BenchmarkResult,
        ):
            markdown_content = (
                f"\n\nBenchmark: Tests with {benchmark_result.name}\n"
            )
            inner_table = Table(
                box=box.SIMPLE_HEAD,
                title=(f"Tests with {benchmark_result.name}"),
                min_width=100,
            )
            rows, columns = _get_sorted_table_from_test_results(
                benchmark_result.test_results,
            )
            for column in columns:
                inner_table.add_column(column)
            for row in rows:
                inner_table.add_row(*row)
            markdown_content += convert_table_to_markdown(columns, rows)
            markdown_content += "\n\n"
            return inner_table, markdown_content

        def _table_for_common_tests(
            benchmark_results: list[BenchmarkResult],
        ):
            markdown_content = (
                f"\nBenchmark: Tests with varying "
                f"{benchmark_group_result.varying_parameter}\n"
            )
            common_tests = {
                test.description for test in benchmark_results[0].test_results
            }
            for benchmark_result in benchmark_results[1:]:
                common_tests.intersection_update(
                    test.description for test in benchmark_result.test_results
                )

            common_test = next(iter(common_tests))
            common_test_table = Table(
                box=box.SIMPLE_HEAD,
                title=(
                    f"Tests with varying "
                    f"{benchmark_group_result.varying_parameter}"
                ),
                min_width=100,
            )
            common_test_results: list[TestResult] = [
                TestResult(
                    description=benchmark_result.name,
                    connectable_results=test_result.connectable_results,
                )
                for benchmark_result in benchmark_results
                for test_result in benchmark_result.test_results
                if test_result.description == common_test
            ]

            rows, columns = _get_sorted_table_from_test_results(
                common_test_results,
            )
            for column in columns:
                common_test_table.add_column(column)
            for row in rows:
                common_test_table.add_row(*row)

            markdown_content += convert_table_to_markdown(
                columns,
                rows,
            )
            markdown_content += "\n\n"
            return common_test_table, markdown_content

        cpu_count = self._report.metadata.system_metadata.get(
            "cpu_count",
            "Not Available",
        )
        cpu_freq = self._report.metadata.system_metadata.get(
            "cpu_freq",
            "Not Available",
        )
        cpu_model = self._report.metadata.system_metadata.get(
            "cpu_model_name",
            "Not Available",
        )
        hostname = self._report.metadata.system_metadata.get(
            "hostname",
            "Not Available",
        )

        benchmark_metadata = (
            f"Benchmark Metadata\n\n"
            f"Runs: {self._report.metadata.num_runs}\n"
            f"Values: {self._report.metadata.num_values}\n"
            f"Warmups: {self._report.metadata.num_warmups}\n\n"
            f"CPU Count: {cpu_count}\n"
            f"CPU Frequency: {cpu_freq}\n"
            f"CPU Model: {cpu_model}\n"
            f"Hostname: {hostname}"
        )

        markdown_content = "# Benchmark Summary\n"
        table = Table(
            Column(
                header="Benchmark Group",
                vertical="middle",
                justify="center",
            ),
            Column(header="Results", justify="center"),
            title="Benchmark Summary",
            caption=benchmark_metadata,
        )

        for (
            benchmark_group_name,
            benchmark_group_result,
        ) in self._report.results.items():
            benchmark_group_path = self._benchmark_group_uri[
                benchmark_group_name
            ]

            markdown_content += f"## Benchmark Group: {benchmark_group_name}\n"
            markdown_content += f"Benchmark File: {benchmark_group_path}\n"
            outer_table = Table(
                box=box.SIMPLE_HEAD,
                caption='File "' + str(benchmark_group_path) + '"',
            )

            if benchmark_group_result.varying_parameter:
                common_tests_table, common_tests_table_markdown = (
                    _table_for_common_tests(
                        benchmark_group_result.benchmark_results,
                    )
                )

                outer_table.add_row(common_tests_table)
                markdown_content += common_tests_table_markdown

            inner_table, inner_table_markdown = _table_for_benchmark_result(
                benchmark_group_result.benchmark_results[-1],
            )

            outer_table.add_row(inner_table)
            markdown_content += inner_table_markdown

            if len(benchmark_group_result.benchmark_results) > 0:
                table.add_row(
                    benchmark_group_name,
                    outer_table,
                )

        markdown_content += "## " + benchmark_metadata

        if self._format == "markdown":
            STDOUT.print(markdown_content)
        elif self._format == "pretty":
            STDOUT.print(table)


@frozen
class Benchmarker:
    _benchmark_groups: Iterable[BenchmarkGroup] = field(
        alias="benchmark_groups",
    )
    _num_runs: int = field(
        alias="runs",
    )
    _num_loops: int = field(
        alias="loops",
    )
    _num_warmups: int = field(
        alias="warmups",
    )
    _num_values: int = field(
        alias="values",
    )
    _num_retries: int = field(
        default=8,
    )

    @classmethod
    def from_default_benchmarks(cls, **kwargs: Any):
        bowtie_dir = Path(__file__).parent
        benchmark_dir = bowtie_dir / "benchmarks"

        if not benchmark_dir.exists():
            raise BenchmarkLoadError("Default Benchmarks not found.")

        return cls._from_dict(
            benchmark_groups=BenchmarkGroup.from_folder(
                benchmark_dir,
            ),
            **kwargs,
        )

    @classmethod
    def from_test_cases(
        cls,
        cases: Iterable[TestCase],
        **kwargs: Any,
    ):
        benchmark_groups = [
            BenchmarkGroup(
                name=case.description,
                description=case.description,
                benchmarks=[
                    Benchmark(
                        name=case.description,
                        schema=case.schema,
                        description=case.description,
                        tests=case.tests,
                    ),
                ],
                uri=None,
                benchmark_type="test_case",
            )
            for case in cases
        ]

        return cls._from_dict(
            benchmark_groups=benchmark_groups,
            **kwargs,
        )

    @classmethod
    def for_keywords(cls, dialect: Dialect, **kwargs: Any):
        bowtie_parent_dir = Path(__file__).parent.parent
        keywords_benchmark_dir = (
            bowtie_parent_dir / "bowtie/benchmarks/keywords"
        )

        if not keywords_benchmark_dir.exists():
            raise BenchmarkLoadError("Keyword Benchmarks Folder not found.")

        keyword_benchmarks = get_benchmark_files(
            benchmark_type="keyword",
            dialect=dialect,
        )

        benchmark_groups: list[BenchmarkGroup] = []
        for keyword_benchmark in keyword_benchmarks:
            relative_folder_path = keyword_benchmark.parent.relative_to(
                bowtie_parent_dir,
            )
            module_name = str(relative_folder_path).replace(os.sep, ".")
            benchmark_group = BenchmarkGroup.from_file(
                keyword_benchmark,
                module_name,
            )
            if (
                benchmark_group is not None
                and dialect in benchmark_group.dialects_supported
            ):
                benchmark_groups.append(
                    benchmark_group,
                )

        return cls._from_dict(
            benchmark_groups=benchmark_groups,
            **kwargs,
        )

    @classmethod
    def for_benchmark_files(
        cls,
        benchmark_files: Iterable[str],
        **kwargs: Any,
    ):
        benchmark_groups: list[BenchmarkGroup] = []
        for benchmark_filename in benchmark_files:
            benchmark_file = Path(benchmark_filename).absolute()
            bowtie_parent_dir = Path(__file__).parent.parent

            benchmark_folder = benchmark_file.parent
            relative_path = benchmark_folder.relative_to(bowtie_parent_dir)

            if not benchmark_file.exists():
                raise BenchmarkLoadError("Benchmark File not found !!")

            module_name = str(relative_path).replace(os.sep, ".")

            benchmark_group = BenchmarkGroup.from_file(
                benchmark_file,
                module=module_name,
            )
            if benchmark_group:
                benchmark_groups.append(benchmark_group)

        return cls._from_dict(
            benchmark_groups=benchmark_groups,
            **kwargs,
        )

    @classmethod
    def from_input(
        cls,
        benchmark: dict[str, Any],
        **kwargs: Any,
    ):
        benchmark["benchmark_type"] = "from_input"
        benchmark_validated(benchmark)
        benchmark_group = BenchmarkGroup.from_dict(
            benchmark,
        )
        return cls._from_dict(benchmark_groups=[benchmark_group], **kwargs)

    @classmethod
    def _from_dict(
        cls,
        benchmark_groups: Iterable[BenchmarkGroup],
        runs: int,
        values: int,
        loops: int,
        warmups: int,
        **kwargs: Any,
    ):
        return cls(
            benchmark_groups=benchmark_groups,
            runs=runs,
            values=values,
            loops=loops,
            warmups=warmups,
        )

    async def start(
        self,
        connectables: Iterable[Connectable],
        dialect: Dialect,
        quiet: bool,
        format: str,
    ):

        acknowledged: Mapping[ConnectableId, ImplementationInfo] = {}
        compatible_connectables: list[Connectable] = []
        incompatible_connectables: list[Connectable] = []

        for connectable in connectables:
            silent_reporter = _report.Reporter(
                write=lambda **_: None,  # type: ignore[reportUnknownArgumentType]
            )
            async with connectable.connect(
                reporter=silent_reporter,
                registry=Direct.from_id("null").registry(),
            ) as implementation:
                if dialect not in implementation.info.dialects:
                    incompatible_connectables.append(connectable)
                    continue
                acknowledged[connectable.to_terse()] = implementation.info
                compatible_connectables.append(connectable)

        with Progress(
            console=STDOUT,
            transient=True,
            disable=quiet,
        ) as progress:
            reporter = BenchmarkReporter(
                report=BenchmarkReport(
                    metadata=BenchmarkMetadata(
                        implementations=acknowledged,
                        dialect=dialect,
                        num_runs=self._num_runs,
                        num_loops=self._num_loops,
                        num_values=self._num_values,
                        num_warmups=self._num_warmups,
                    ),
                ),
                quiet=quiet,
                format=format,
                progress_bar=progress,
            )
            reporter.started()
            reporter.report_incompatible_connectables(
                incompatible_connectables,
                dialect,
            )

            if not compatible_connectables:
                reporter.no_compatible_connectables()
                return
            some_benchmark_ran = False
            some_benchmark_ran_successfully = False
            for benchmark_group in self._benchmark_groups:
                if dialect not in benchmark_group.dialects_supported:
                    if not quiet:
                        STDOUT.log(
                            f"Skipping Benchmark Group: {benchmark_group.name}"
                            f" as it does not support dialect"
                            f" {dialect.serializable()}",
                        )
                    continue
                (
                    benchmark_started,
                    benchmark_group_finished,
                ) = reporter.running_benchmark_group(
                    benchmark_group,
                )
                for benchmark in benchmark_group.benchmarks:
                    if benchmark.dialect and benchmark.dialect != dialect:
                        if not quiet:
                            STDOUT.log(
                                f"Skipping {benchmark.name} as it does not "
                                f"support dialect {dialect.serializable()}",
                            )
                        continue
                    test_started, benchmark_finished = benchmark_started(
                        benchmark.name,
                        benchmark.description,
                    )
                    tests = benchmark.tests
                    for test in tests:
                        (
                            got_connectable_result,
                            test_finished,
                        ) = test_started(test.description)
                        benchmark_case = benchmark.benchmark_with_diff_tests(
                            tests=[test],
                        )
                        for connectable in compatible_connectables:
                            run_needed, retries_allowed = (
                                True,
                                self._num_retries,
                            )
                            output = ""
                            while run_needed:
                                with tempfile.NamedTemporaryFile(
                                    delete=True,
                                ) as fp:
                                    try:
                                        output = await self._run_benchmark(
                                            benchmark_case,
                                            dialect,
                                            connectable,
                                            fp.name,
                                        )
                                    except BowtieRunError as err:
                                        got_connectable_result(
                                            connectable.to_terse(),
                                            "Errored",
                                            0,
                                            [1e9]
                                            * self._num_runs
                                            * self._num_values,
                                            errored=True,
                                        )
                                        if not quiet:
                                            STDOUT.log(err)

                                        run_needed = False
                                        some_benchmark_ran = True

                                    if run_needed:
                                        lines = (
                                            Path(
                                                fp.name,
                                            )
                                            .read_text()
                                            .splitlines()
                                        )

                                        measured_time_values = [
                                            float(line) / 1e9 for line in lines
                                        ]

                                        run_needed = got_connectable_result(
                                            connectable.to_terse(),
                                            output,
                                            retries_allowed,
                                            measured_time_values,
                                        )

                                        some_benchmark_ran = True
                                        some_benchmark_ran_successfully = True

                                        if not retries_allowed:
                                            run_needed = False

                                        if run_needed:
                                            retries_allowed -= 1

                        test_finished()

                    benchmark_finished()

                benchmark_group_finished()

            if not some_benchmark_ran and not quiet:
                STDOUT.log("No Benchmark ran.")

            if not some_benchmark_ran_successfully:
                raise BowtieRunError("No Benchmark Ran Successfully!", None)

            reporter.finished()

    async def _run_benchmark(
        self,
        benchmark: Benchmark,
        dialect: Dialect,
        connectable: Connectable,
        time_output_file: str,
    ) -> Any:
        benchmark_name = f"{benchmark.name}::{benchmark.tests[0].description}"

        benchmark_dict = benchmark.serializable()
        benchmark_dict["schema"]["$schema"] = str(dialect.uri)
        if "name" in benchmark_dict:
            benchmark_dict.pop("name")

        if "dialect" in benchmark_dict:
            benchmark_dict.pop("dialect")

        with tempfile.NamedTemporaryFile(
            delete=True,
        ) as fp:
            path = Path(fp.name)
            path.write_text(json.dumps(benchmark_dict))
            return await self._pyperf_benchmark_command(
                "bowtie",
                "run",
                "-i",
                connectable.to_terse(),
                "-D",
                dialect.serializable(),
                fp.name,
                time_output_file=time_output_file,
                connectable_id=connectable.to_terse(),
                name=benchmark_name,
            )

    @staticmethod
    async def _run_subprocess(
        *cmd: str,
        env: dict[str, Any] = {},
    ):
        env = {**os.environ, **env}
        process = await asyncio.create_subprocess_exec(
            *cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=env,
        )
        stdout, stderr = await process.communicate()
        return stdout, stderr

    async def _pyperf_benchmark_command(
        self,
        *benchmark_cmd: str,
        time_output_file: str,
        connectable_id: ConnectableId,
        name: str,
    ):
        stdout_fd = "1"

        output, err = await self._run_subprocess(
            "pyperf",
            "command",
            "--pipe",
            stdout_fd,
            "--copy-env",
            "--processes",
            str(self._num_runs),
            "--values",
            str(self._num_values),
            "--warmups",
            str(self._num_warmups),
            "--loops",
            str(self._num_loops),
            "--name",
            name,
            *benchmark_cmd,
            env={
                "TIME_OUTPUT_FILE": time_output_file,
            },
        )
        if err:
            _, inner_err = await self._run_subprocess(
                *benchmark_cmd,
            )
            if inner_err:
                raise BowtieRunError(inner_err.decode(), connectable_id)
            raise PyperfError(err.decode())

        return output



from pathlib import Path

from url.url import URL

from bowtie._benchmarks import Benchmark, BenchmarkGroup
from bowtie._core import Dialect


def get_benchmark():
    name = "contains"
    benchmark_type = "keyword"
    description = "A benchmark for validation of the `contains` keyword."
    max_array_length = 100000
    varying_parameter = "Array length"

    array_length = 1000
    benchmarks = []
    while array_length <= max_array_length:

        start = [37] + [0] * (array_length - 1)
        middle = [0] * (array_length // 2) + [37] + [0] * (array_length // 2)
        end = [0] * (array_length - 1) + [37]
        invalid = [0] * array_length

        tests = (
            [
                dict(description="Empty array", instance=[]),
                dict(description="Beginning of array", instance=start),
                dict(description="Middle of array", instance=middle),
                dict(description="End of array", instance=end),
                dict(description="Invalid array", instance=invalid),
            ]
            if array_length == max_array_length
            else [
                dict(description="Middle of array", instance=middle),
            ]
        )

        benchmarks.append(
            Benchmark.from_dict(
                name=f"Array length - {array_length}",
                description=(
                    "Validating contains keyword over an array "
                    f"of length {array_length}"
                ),
                schema={
                    "type": "array",
                    "contains": {"const": 37},
                },
                tests=tests,
            ),
        )
        array_length *= 10

    return BenchmarkGroup(
        name=name,
        benchmark_type=benchmark_type,
        description=description,
        dialects_supported=[
            Dialect.from_str("https://json-schema.org/draft/2020-12/schema"),
            Dialect.from_str("https://json-schema.org/draft/2019-09/schema"),
            Dialect.from_str("http://json-schema.org/draft-07/schema#"),
            Dialect.from_str("http://json-schema.org/draft-06/schema#"),
        ],
        benchmarks=benchmarks,
        uri=URL.parse(Path(__file__).absolute().as_uri()),
        varying_parameter=varying_parameter,
    )



from pathlib import Path

from url.url import URL

from bowtie._benchmarks import Benchmark, BenchmarkGroup
from bowtie._core import Dialect


def get_benchmark():
    name = "uniqueItems"
    benchmark_type = "keyword"
    description = (
        "A benchmark for measuring performance of the "
        "implementation for the uniqueItems keyword."
    )
    max_array_length = 200000
    varying_parameter = "Array length"

    array_length = 2000
    benchmarks = []
    while array_length <= max_array_length:
        first_two_duplicate = [1, 1, *list(range(2, array_length - 2))]
        middle_two_duplicate = [
            *list(range(array_length // 2)),
            -1,
            -1,
            *list(range(array_length // 2, array_length)),
        ]
        last_two_duplicate = [*list(range(2, array_length - 2)), 1, 1]
        valid = list(range(array_length))

        tests = (
            [
                dict(
                    description="First Two Duplicate",
                    instance=first_two_duplicate,
                ),
                dict(
                    description="Middle Two Duplicate",
                    instance=middle_two_duplicate,
                ),
                dict(
                    description="Last Two Duplicate",
                    instance=last_two_duplicate,
                ),
                dict(description="Valid", instance=valid),
            ]
            if array_length == max_array_length
            else [
                dict(description="Valid", instance=valid),
            ]
        )

        benchmarks.append(
            Benchmark.from_dict(
                name=f"Array length - {array_length}",
                description=(
                    f"Validating the `uniqueItems` keyword over array of length {array_length}."
                ),
                schema=dict(
                    uniqueItems=True,
                ),
                tests=tests,
            ),
        )

        array_length *= 10

    return BenchmarkGroup(
        name=name,
        benchmark_type=benchmark_type,
        description=description,
        dialects_supported=[
            Dialect.from_str("https://json-schema.org/draft/2020-12/schema"),
            Dialect.from_str("https://json-schema.org/draft/2019-09/schema"),
            Dialect.from_str("http://json-schema.org/draft-07/schema#"),
            Dialect.from_str("http://json-schema.org/draft-06/schema#"),
            Dialect.from_str("http://json-schema.org/draft-04/schema#"),
            Dialect.from_str("http://json-schema.org/draft-03/schema#"),
        ],
        benchmarks=benchmarks,
        uri=URL.parse(Path(__file__).absolute().as_uri()),
        varying_parameter=varying_parameter,
    )



from bowtie._benchmarks import Benchmark


def get_benchmark():
    return Benchmark.from_dict(
        name="benchmark",
        schema={
            "type": "object",
        },
        description="benchmark",
        tests=[
            {
                "description": "test",
                "instance": {
                    "a": "b",
                },
            },
        ],
    )



from pathlib import Path
import json

import pytest

from bowtie import HOMEPAGE, REPO, _benchmarks, benchmarks
from bowtie._core import Dialect, ImplementationInfo, TestCase
from bowtie._direct_connectable import Direct

validators = Direct.from_id("python-jsonschema").registry()

benchmark_report_validator = validators.for_uri(
    "tag:bowtie.report,2024:benchmark_report",
)
benchmark_validator = validators.for_uri("tag:bowtie.report,2024:benchmarks")


def collect_benchmarks(directory):
    return sorted(
        path
        for path in directory.iterdir()
        if path.suffix == ".json"
        or path.suffix == ".py"
        and path.name != "__init__.py"
    )


BENCHMARKS_DIR = Path(benchmarks.__file__).parent
BENCHMARK_PATHS = collect_benchmarks(BENCHMARKS_DIR)
KEYWORD_BENCHMARK_PATHS = collect_benchmarks(BENCHMARKS_DIR / "keywords")

DIRECT_CONNECTABLE = "python-jsonschema"


@pytest.fixture()
def valid_single_benchmark():
    from bowtie.tests.benchmarks import valid_single_benchmark

    return valid_single_benchmark.get_benchmark()


@pytest.fixture()
def valid_benchmark_group(valid_single_benchmark):
    from bowtie.tests.benchmarks import valid_benchmark_group

    return valid_benchmark_group.get_benchmark()


@pytest.fixture()
def invalid_benchmark():
    from bowtie.tests.benchmarks import invalid_benchmark

    return invalid_benchmark.get_benchmark()


@pytest.fixture()
def benchmarker_run_args():
    return {
        "runs": 1,
        "loops": 1,
        "warmups": 1,
        "values": 2,
    }


def _validate_benchmark_file(file, module):
    data = _benchmarks._load_benchmark_group_from_file(file, module)
    assert benchmark_validator.is_valid(data.serializable())


class TestBenchmarkFormat:

    def test_validate_single_benchmark(self, valid_single_benchmark):
        assert benchmark_validator.is_valid(
            valid_single_benchmark.serializable(),
        )

    def test_validate_benchmark_group(self, valid_benchmark_group):
        assert benchmark_validator.is_valid(
            valid_benchmark_group.serializable(),
        )

    @pytest.mark.parametrize("path", BENCHMARK_PATHS, ids=str)
    def test_validate_default_benchmark_format(self, path):
        benchmark_module = "bowtie.benchmarks"
        _validate_benchmark_file(path, benchmark_module)

    @pytest.mark.parametrize("path", KEYWORD_BENCHMARK_PATHS, ids=str)
    def test_validate_keyword_benchmark_format(self, path):
        benchmark_module = "bowtie.benchmarks.keywords"
        _validate_benchmark_file(path, benchmark_module)


class TestLoadBenchmark:

    def test_benchmark_set_dialect(self, valid_single_benchmark):
        benchmark_json = valid_single_benchmark.serializable()
        benchmark_json["schema"]["$schema"] = Dialect.latest().serializable()
        benchmark_with_explicit_dialect = _benchmarks.Benchmark.from_dict(
            **benchmark_json,
        ).maybe_set_dialect_from_schema()
        assert benchmark_with_explicit_dialect.dialect == Dialect.latest()

    def test_load_benchmark_with_diff_tests(self, valid_single_benchmark):
        benchmark = valid_single_benchmark.benchmark_with_diff_tests(
            tests=valid_single_benchmark.tests * 10,
        )
        assert benchmark_validator.is_valid(benchmark.serializable())

    def test_load_single_benchmark_group_from_dict(
        self,
        valid_single_benchmark,
    ):
        benchmark = valid_single_benchmark.serializable()
        benchmark["benchmark_type"] = "test"
        benchmark_group = _benchmarks.BenchmarkGroup.from_dict(benchmark)

        assert benchmark_validator.is_valid(benchmark_group.serializable())

    def test_load_benchmark_group_from_dict(self, valid_benchmark_group):
        benchmark_json = valid_benchmark_group.serializable()
        benchmark_group = _benchmarks.BenchmarkGroup.from_dict(
            benchmark_json,
            uri=valid_benchmark_group.uri,
        )

        serializable = benchmark_group.serializable()

        assert benchmark_validator.is_valid(serializable)
        assert valid_benchmark_group.serializable() == serializable

    def test_load_single_benchmark_group_from_json(
        self,
        tmp_path,
        valid_single_benchmark,
    ):
        tmp_path = tmp_path / "test_file.json"
        single_benchmark_json = valid_single_benchmark.serializable()
        single_benchmark_json["benchmark_type"] = "test"
        tmp_path.write_text(json.dumps(single_benchmark_json))
        benchmark_group = _benchmarks.BenchmarkGroup.from_file(tmp_path)
        assert benchmark_validator.is_valid(benchmark_group.serializable())

    def test_load_benchmark_group_from_json(
        self,
        tmp_path,
        valid_benchmark_group,
    ):
        tmp_path = tmp_path / "test_file.json"

        benchmark_group_json = valid_benchmark_group.serializable()
        benchmark_group_json["uri"] = str(tmp_path.absolute().as_uri())
        tmp_path.write_text(json.dumps(benchmark_group_json))

        loaded_benchmark_group = _benchmarks.BenchmarkGroup.from_file(tmp_path)

        assert (
            benchmark_validator.validated(
                loaded_benchmark_group.serializable(),
            )
            == benchmark_group_json
        )

    def test_load_benchmark_groups_from_folder(self):
        benchmark_groups = _benchmarks.BenchmarkGroup.from_folder(
            Path(__file__).parent / "benchmarks",
            module="bowtie.tests.benchmarks",
        )
        valid_benchmarks_for_test = 3
        valid_count = 0

        for benchmark_group in benchmark_groups:
            assert benchmark_validator.is_valid(benchmark_group.serializable())
            valid_count += 1

        assert valid_count == valid_benchmarks_for_test


class TestBenchmarker:

    def test_default_benchmarker(self, benchmarker_run_args):
        _benchmarks.Benchmarker.from_default_benchmarks(**benchmarker_run_args)

    @pytest.mark.parametrize(
        "dialect",
        Dialect.known(),
        ids=lambda param: param.short_name,
    )
    def test_keywords_benchmarker(self, dialect, benchmarker_run_args):
        _benchmarks.Benchmarker.for_keywords(dialect, **benchmarker_run_args)

    def test_test_cases_benchmarker(
        self,
        valid_single_benchmark,
        benchmarker_run_args,
    ):
        test_case = TestCase(
            description=valid_single_benchmark.description,
            schema=valid_single_benchmark.schema,
            tests=valid_single_benchmark.tests,
        )

        _benchmarks.Benchmarker.from_test_cases(
            [test_case],
            **benchmarker_run_args,
        )

    def test_input_benchmarker(
        self,
        valid_single_benchmark,
        benchmarker_run_args,
    ):
        _benchmarks.Benchmarker.from_input(
            valid_single_benchmark.serializable(),
            **benchmarker_run_args,
        )


class TestBenchmarkReport:

    def test_merge_benchmark_reports(
        self,
    ):
        benchmark_metadata = _benchmarks.BenchmarkMetadata(
            implementations={
                "foo": ImplementationInfo(
                    name="foo",
                    language="blub",
                    homepage=HOMEPAGE,
                    issues=REPO / "issues",
                    source=REPO,
                    dialects=frozenset([Dialect.latest()]),
                ),
            },
            dialect=Dialect.latest(),
            num_runs=1,
            num_loops=1,
            num_values=1,
            num_warmups=1,
        )

        benchmark_results = [
            _benchmarks.BenchmarkResult(
                name="benchmark1",
                description="test",
                test_results=[
                    _benchmarks.TestResult(
                        description="test",
                        connectable_results=[
                            _benchmarks.ConnectableResult(
                                connectable_id="foo",
                                duration=1.0,
                                values=[1.0],
                                errored=False,
                            ),
                        ],
                    ),
                ],
            ),
        ]
        benchmark_report1 = _benchmarks.BenchmarkReport(
            metadata=benchmark_metadata,
            results={
                "benchmark_group_1": _benchmarks.BenchmarkGroupResult(
                    name="benchmark_group_1",
                    description="test",
                    benchmark_type="test",
                    benchmark_results=benchmark_results,
                ),
            },
        )
        benchmark_report2 = _benchmarks.BenchmarkReport(
            metadata=benchmark_metadata,
            results={
                "benchmark_group_2": _benchmarks.BenchmarkGroupResult(
                    name="benchmark_group_2",
                    description="test",
                    benchmark_type="test",
                    benchmark_results=benchmark_results,
                ),
            },
        )
        merged_report = _benchmarks.BenchmarkReport.merge(
            [benchmark_report1, benchmark_report2],
        )
        assert len(merged_report.results.items()) == len(
            benchmark_report1.results.items(),
        ) + len(benchmark_report2.results.items())
        assert "benchmark_group_1" in merged_report.results
        assert "benchmark_group_2" in merged_report.results



from pathlib import Path
import uuid

from url.url import URL

from bowtie._benchmarks import Benchmark, BenchmarkGroup
from bowtie._core import Dialect


def get_benchmark():

    name = "minContains"
    benchmark_type = "keyword"
    description = (
        "A benchmark for measuring performance of the implementation "
        "for the minContains keyword."
    )
    max_array_length = 100000
    varying_parameter = "Array length"

    array_length = 1000
    benchmarks = []
    while array_length <= max_array_length:
        array = [uuid.uuid4().hex for _ in range(array_length)]

        both_at_first = [1, 1] + array[:-2]
        both_at_middle = (
            array[1 : array_length // 2]
            + [1, 1]
            + array[array_length // 2 : -1]
        )
        both_at_last = array[:-2] + [1, 1]
        invalid = array

        tests = (
            [
                dict(description="Both at First", instance=both_at_first),
                dict(
                    description="Both at Middle",
                    instance=both_at_middle,
                ),
                dict(description="Both at Last", instance=both_at_last),
                dict(description="Invalid", instance=invalid),
            ]
            if array_length == max_array_length
            else [
                dict(
                    description="Both at Middle",
                    instance=both_at_middle,
                ),
            ]
        )

        benchmarks.append(
            Benchmark.from_dict(
                name=f"minContains_{array_length}",
                description=(
                    "A benchmark for validation of the `minContains` keyword."
                ),
                schema={
                    "type": "array",
                    "contains": {"type": "integer"},
                    "minContains": 2,
                },
                tests=tests,
            ),
        )
        array_length *= 10

    return BenchmarkGroup(
        name=name,
        benchmark_type=benchmark_type,
        description=description,
        dialects_supported=[
            Dialect.from_str("https://json-schema.org/draft/2020-12/schema"),
            Dialect.from_str("https://json-schema.org/draft/2019-09/schema"),
        ],
        benchmarks=benchmarks,
        uri=URL.parse(Path(__file__).absolute().as_uri()),
        varying_parameter=varying_parameter,
    )



"""
A collection of small mini-fake implementations with deterministic behavior.

Useful for testing Bowtie's own internal behavior.
"""

from url import URL

from bowtie import HOMEPAGE, REPO
from bowtie._commands import StartedDialect
from bowtie._core import Dialect, Link
from bowtie._direct_connectable import DirectImplementation, null

ARBITRARILY_INVALID = ExceptionGroup("Not valid!", [ZeroDivisionError()])


def fake(**kwargs):
    return DirectImplementation.from_callable(
        **{
            "homepage": HOMEPAGE,
            "source": REPO,
            "issues": REPO / "issues",
            "version": "v1.0.0",
            "dialects": frozenset(Dialect.known()),
            **kwargs,
        },
    )


@fake()
def always_invalid(dialect: Dialect):
    """
    Always considers any instance invalid.

    For the inverse, see `Direct.null`.
    """
    return lambda schema, registry: lambda instance: ARBITRARILY_INVALID


@fake()
def always_wrong(dialect: Dialect):
    """
    Tries (naively) to always get the absolute wrong answer.
    """
    return lambda schema, registry: lambda instance: (
        naively_incorrect(schema, registry, instance)
    )


def only_supports(dialect: str):
    """
    Claims to only support the dialect provided in its (sole) parameter.

    Use this by passing a connectable parameter, e.g. via
    ``direct:miniatures:only_supports,dialect=draft4``.

    The validity result of instances should not be relied on.
    """
    return fake(
        name="only_supports",
        dialects=frozenset([Dialect.by_short_name()[dialect]]),
    )(null)()


@fake()
def incorrectly_claims_draft7(dialect: Dialect):
    """
    Claims to support Draft 7 but returns all wrong results for it.

    Otherwise attempts to crudely pass the smoke test.
    """
    if dialect == dialect.by_short_name()["draft7"]:
        return lambda schema, registry: lambda instance: (
            naively_correct(schema, registry)(instance)
            if registry
            else naively_incorrect(schema, registry, instance)
        )
    return naively_correct


@fake(language="javascript")
def fake_javascript(dialect: Dialect):
    """
    Claims to be written in Javascript.

    The validity result of instances should not be relied on.
    """
    return lambda schema, registry: lambda instance: None


@fake()
def passes_smoke(dialect: Dialect):
    """
    An implementation which crudely passes `bowtie smoke`.
    """
    return naively_correct


@fake()
def no_registry_support(dialect: Dialect):
    """
    An implementation which tries really hard but fails at referecing.
    """
    return lambda schema, registry: (
        naively_correct(schema, registry)
        if not registry
        else lambda instance: None
    )


@fake(implicit_dialect_response=StartedDialect(ok=False))
def no_implicit_dialect_support(dialect: Dialect):
    """
    An implementation which does not support implicit dialects in schemas.

    The validity result of instances should not be relied on.
    """
    return lambda schema, registry: lambda instance: None


@fake(
    homepage=URL.parse("urn:example"),
    issues=URL.parse("urn:example"),
    source=URL.parse("urn:example"),
    dialects=frozenset([Dialect.by_short_name()["draft7"]]),
    links=[
        Link(description="foo", url=URL.parse("urn:example:foo")),
        Link(description="bar", url=URL.parse("urn:example:bar")),
    ],
)
def links(dialect: Dialect):
    """
    An implementation which declares some additional links.

    The validity result of instances should not be relied on.
    """
    return lambda schema, registry: lambda instance: None


def has_bugs_by_versions(version: str):
    """
    A buggy implementation whose behaviour changes based on its versions.

    Use this by passing a connectable parameter, e.g. via
    ``direct:miniatures:has_bugs_by_versions,version=1.0``.

    The validity result of instances should not be relied on.
    """
    if version == "1.0":
        return fake(
            name="buggy",
            version=version,
        )(
            lambda dialect: (
                (lambda schema, registry: lambda instance: ARBITRARILY_INVALID)
                if dialect == Dialect.by_alias()["2020"]
                else (
                    (lambda schema, registry: lambda instance: None)
                    if dialect == Dialect.by_alias()["2019"]
                    else (lambda schema, registry: lambda instance: None)
                )
            ),
        )()
    elif version == "2.0":
        return fake(
            name="buggy",
            version=version,
        )(
            lambda dialect: (
                (lambda schema, registry: lambda instance: None)
                if dialect == Dialect.by_alias()["2020"]
                else (
                    (
                        lambda schema, registry: (
                            lambda instance: ARBITRARILY_INVALID
                        )
                    )
                    if dialect == Dialect.by_alias()["2019"]
                    else (lambda schema, registry: lambda instance: None)
                )
            ),
        )()
    return fake(name="buggy")(null)()


def naively_correct(schema, registry):
    """
    The naivest implementation which tries to pass a smoke test.
    """
    ref = schema.get("$ref")
    if ref is not None:
        schema = registry.contents(ref)

    def validate(instance):
        if "not" in schema or "disallow" in schema:
            return ARBITRARILY_INVALID

        pytype = dict(string=str).get(schema.get("type"))
        if pytype is not None and not isinstance(instance, pytype):
            return ARBITRARILY_INVALID

    return validate


def naively_incorrect(schema, registry, instance):
    """
    Naively returns the wrong answer.
    """
    result = naively_correct(schema, registry)(instance)
    return ARBITRARILY_INVALID if result is None else None



"""
Validator registries, likely for eventual upstreaming into referencing.
"""

from __future__ import annotations

from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from attrs import evolve, field, frozen
from referencing.jsonschema import EMPTY_REGISTRY, Schema, SchemaRegistry

if TYPE_CHECKING:
    from referencing.jsonschema import SchemaResource
    from url import URL


E_co = TypeVar("E_co", bound=Exception, covariant=True)
Validate = Callable[[Any], ExceptionGroup[E_co] | None]
SchemaCompiler = Callable[[Schema, SchemaRegistry], Validate[E_co]]


class Invalid(ExceptionGroup[E_co]):
    """
    An instance is not valid under a schema.
    """

    # TODO: Don't require a message? Or what should the general message be?
    #       What about lazy error production? Can we delay asking for all
    #           errors here?


class UnexpectedlyValid(Exception):
    """
    An instance which was expected to be invalid just isn't.
    """


@frozen
class Validator(Generic[E_co]):
    """
    A compiled schema, ready to validate instances.
    """

    validate: Validate[E_co]
    _registry: ValidatorRegistry[E_co] = field(alias="registry")

    def __rmatmul__(
        self,
        resources: SchemaResource | Iterable[SchemaResource],
    ) -> Validator[E_co]:
        return evolve(self, registry=resources @ self._registry)

    def is_valid(self, instance: Any):
        return self.validate(instance) is None

    def validated(self, instance: Any):
        exception = self.validate(instance)
        if exception is not None:
            raise exception
        return instance

    def invalidated(self, instance: Any):
        if self.is_valid(instance):
            raise UnexpectedlyValid(instance)
        return instance


@frozen
class ValidatorRegistry(Generic[E_co]):

    _compile: SchemaCompiler[E_co] = field(alias="compile")
    _registry: SchemaRegistry = field(default=EMPTY_REGISTRY, alias="registry")

    def __rmatmul__(
        self,
        resources: SchemaResource | Iterable[SchemaResource],
    ) -> ValidatorRegistry[E_co]:
        return evolve(self, registry=resources @ self._registry)

    def schema(self, uri: URL) -> Schema:
        """
        Return the schema identified by the given URI.
        """
        return self._registry.resolver().lookup(str(uri)).contents

    def for_uri(self, uri: URL) -> Validator[E_co]:
        """
        Return a `Validator` using the schema at the given URI.
        """
        return self.for_schema(self.schema(uri))

    def for_schema(self, schema: Schema) -> Validator[E_co]:
        """
        Return a `Validator` using the given schema.
        """
        validate: Validate[E_co] = self._compile(schema, self._registry)
        return Validator(validate=validate, registry=self)



from pathlib import Path
import uuid

from url.url import URL

from bowtie._benchmarks import Benchmark, BenchmarkGroup
from bowtie._core import Dialect


def get_benchmark():
    name = "additionalProperties"
    benchmark_type = "keyword"
    description = (
        "A benchmark for measuring performance of the "
        "implementation for the additionalProperties keyword."
    )
    max_array_length = 100000
    varying_parameter = "Array length"

    array_length = 1000
    benchmarks: list[Benchmark] = []
    while array_length <= max_array_length:
        allowed_properties = [
            uuid.uuid4().hex for _ in range(max_array_length - 1)
        ]

        middle_index = len(allowed_properties) // 2
        invalid_at_first = [uuid.uuid4().hex, *allowed_properties]
        invalid_at_middle = (
            allowed_properties[:middle_index]
            + [uuid.uuid4().hex]
            + allowed_properties[middle_index:]
        )
        invalid_at_last = [*allowed_properties, uuid.uuid4().hex]
        valid = allowed_properties

        tests = (
            [
                dict(
                    description="Invalid at first",
                    instance=_format_properties_as_instance(
                        invalid_at_first,
                    ),
                ),
                dict(
                    description="Invalid at middle",
                    instance=_format_properties_as_instance(
                        invalid_at_middle,
                    ),
                ),
                dict(
                    description="Invalid at last",
                    instance=_format_properties_as_instance(
                        invalid_at_last,
                    ),
                ),
                dict(
                    description="Valid",
                    instance=_format_properties_as_instance(valid),
                ),
            ]
            if array_length == max_array_length
            else [
                dict(
                    description="Valid",
                    instance=_format_properties_as_instance(valid),
                ),
            ]
        )

        benchmarks.append(
            Benchmark.from_dict(
                name=f"Array length - {array_length}",
                description=(
                    f"Validating additionalProperties keyword over array of length {array_length}"
                ),
                schema=dict(
                    properties={key: {} for key in allowed_properties},
                    additionalProperties=False,
                ),
                tests=tests,
            ),
        )

        array_length *= 10

    return BenchmarkGroup(
        name=name,
        benchmark_type=benchmark_type,
        dialects_supported=[
            Dialect.from_str("https://json-schema.org/draft/2020-12/schema"),
            Dialect.from_str("https://json-schema.org/draft/2019-09/schema"),
            Dialect.from_str("http://json-schema.org/draft-07/schema#"),
            Dialect.from_str("http://json-schema.org/draft-06/schema#"),
            Dialect.from_str("http://json-schema.org/draft-04/schema#"),
            Dialect.from_str("http://json-schema.org/draft-03/schema#"),
        ],
        description=description,
        benchmarks=benchmarks,
        varying_parameter=varying_parameter,
        uri=URL.parse(Path(__file__).absolute().as_uri()),
    )


def _format_properties_as_instance(properties):
    return {key: 1000 for key in properties}



"""
Errors. Oh no!

Bowtie's communication with an implementation breaks down into a sequence
of steps roughly listed below, along with a possible exception that
connectables and/or harnesses can raise if the corresponding step fails.

    * connect to the implementation (fails via `CannotConnect` or via
    `NoSuchImplementation` if we seem sure the implementation doesn't exist)
    * send a first request which retrieves metadata about the implementation
      (fails via `StartupFailed` or `VersionMismatch`)
    * send sequences of test cases, collecting results
    * close the connection

At any given point in time after connecting, `InvalidResponse` or
`ProtocolError` may be raised to indicate that the harness is not properly
speaking Bowtie's IO protocol, and `GotStderr` may be raised to indicate the
harness has spewed data to standard error which likely means it will not return
a response to any current request. All of these typically are harness bugs, as
the harness should generally catch implementation crashes and respond within
Bowtie's protocol.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from attrs import frozen
from diagnostic import DiagnosticError, DiagnosticWarning

if TYPE_CHECKING:
    from collections.abc import Iterable

    from rich.console import Console, ConsoleOptions, RenderResult

    from bowtie._connectables import ConnectableId
    from bowtie._core import Dialect, Implementation


#: The current version of Bowtie's IO protocol.
_PROTOCOL_VERSION = 1


@frozen
class NoSuchImplementation(Exception):
    """
    An implementation with the given name does not exist.

    This means we don't know what implementation is intended, either because
    Bowtie doesn't support it yet or because its name contains a typo.
    """

    id: ConnectableId

    def __rich__(self):
        return DiagnosticError(
            code="no-such-implementation",
            message=f"{self.id!r} is not a known Bowtie implementation.",
            causes=[],
            hint_stmt=(
                "Check Bowtie's supported list of implementations "
                "to ensure you have the name correct. "
                "If you are developing a new harness, ensure you have "
                "built and tagged it properly."
            ),
        )


@frozen
class CannotConnect(Exception):
    """
    We could not connect to the desired implementation.

    It may exist, but this exception may also be used to indicate we could not
    even check whether that was the case for some connectable-specific reason.

    In particular, either the implementation doesn't support connection via
    the connectable used, or it does but is unreachable in a way which will
    be connectable-specific (e.g. general networking issues).
    """

    kind: str
    id: ConnectableId
    hint: str | None = None

    def __rich__(self):
        cause = self.__cause__ or self.__context__
        causes = [str(cause)] if cause is not None else []

        fqid = f"{self.kind}:{self.id}"
        return DiagnosticError(
            code="cannot-connect",
            message=f"Couldn't connect to {fqid!r}.",
            causes=causes,
            hint_stmt=self.hint,
        )


@frozen
class StartupFailed(Exception):
    """
    We found to an implementation but could not get it to start.
    """

    id: ConnectableId
    stderr: str = ""
    data: Any = None

    def __rich_console__(
        self,
        console: Console,
        options: ConsoleOptions,
    ) -> RenderResult:
        cause = self.__cause__
        if cause is None:
            root_causes, hint = [], (
                "If you are developing a new harness, check if stderr "
                "(shown below) contains harness-specific information "
                "which can help. Otherwise, you may have an issue with your "
                "local container setup (podman, docker, etc.)."
            )
        else:
            root_causes: Iterable[Exception] = getattr(
                cause.__cause__,
                "exceptions",
                [],
            )
            hint = (
                "The harness sent an invalid response for Bowtie's protocol. "
                "Details for what was wrong are above. If you are developing "
                "support for a new harness you should address them, otherwise "
                "if you are not, this is a bug in Bowtie's harness for this "
                "implementation! File an issue on Bowtie's issue tracker."
            )

        yield DiagnosticError(
            code="startup-failed",
            message=f"{self.id!r} failed to start.",
            causes=[str(exc) for exc in root_causes],
            hint_stmt=hint,
        )
        if self.stderr:
            from rich.panel import Panel

            yield Panel(self.stderr, title="stderr")


@frozen
class VersionMismatch(Exception):
    """
    The wrong protocol version was returned from an implementation.
    """

    got: int
    expected: int = _PROTOCOL_VERSION

    def __str__(self) -> str:
        return (
            f"Expected to speak version {self.expected} of the Bowtie "
            f"protocol but the implementation sent {self.got}."
        )

    @classmethod
    def check(cls, got: int) -> None:
        """
        Complain if we are speaking the wrong protocol version.
        """
        if got != _PROTOCOL_VERSION:
            raise cls(got=got)


@frozen
class ProtocolError(Exception):
    """
    An invalid request or response was sent.
    """


@frozen
class InvalidResponse(Exception):
    """
    An invalid response was sent by a harness.
    """

    # TODO: Combine with ProtocolError?

    contents: str


@frozen
class UnsupportedDialect(Exception):
    """
    An implementation does not support the dialect which is to be spoken.
    """

    implementation: Implementation
    dialect: Dialect

    def __rich__(self):
        supports = ", ".join(
            each.pretty_name
            for each in sorted(self.implementation.info.dialects, reverse=True)
        )
        return DiagnosticWarning(
            code="unsupported-dialect",
            message=(
                f"{self.implementation.id!r} does not "
                f"support {self.dialect.pretty_name}."
            ),
            causes=[],
            hint_stmt=None,
            note_stmt=f"its supported dialects are: {supports}\n",
        )


@frozen
class DialectError(Exception):
    """
    We tried to start sending test cases but encountered an unknown error.

    It probably ain't our fault, the implementation blew up.
    """

    implementation: Implementation
    dialect: Dialect
    stderr: bytes

    def __rich__(self):
        return DiagnosticError(
            code="dialect-error",
            message=(
                f"{self.implementation.id!r} failed as we were beginning to "
                f"send {self.dialect.pretty_name} tests."
            ),
            causes=[],
            hint_stmt=(
                "The error may be transient, so you may want to try again. "
                "If it does not appear to be, it may indicate a bug both in "
                "the implementation as well as in Bowtie's test harness. "
                "Bowtie aims not to propagate these bugs in a way that they "
                "cause crashes, so at very least you should open a bug "
                "(with reproducer) on Bowtie's issue board."
            ),
        )


@frozen
class GotStderr(Exception):
    """
    An implementation sent data on standard error.

    We were trying to communicate with it (via Bowtie's protocol), but the
    implementation has likely encountered some unexpected error.

    It may have crashed.

    Implementations of the `Connection` protocol should raise this exception
    when they detect this kind of out-of-band error (in whatever concrete
    connection-specific mechanism indicates this has occurred).
    """

    stderr: bytes



"""
A meta-validator for the JSON Schema specification.
"""

from url import URL

HOMEPAGE = URL.parse("https://bowtie.report/")
DOCS = URL.parse("https://docs.bowtie.report/")

GITHUB = URL.parse("https://github.com/")
ORG_NAME = "bowtie-json-schema"
ORG = GITHUB / ORG_NAME
REPO = ORG / "bowtie"



"""
Test Bowtie's schemas for proper functionality.
"""

import pytest

from bowtie._direct_connectable import Direct

TEST = {
    "description": "a test",
    "instance": 37,
}
ANOTHER_TEST = {
    "description": "another test",
    "instance": "foo",
}


@pytest.mark.parametrize(
    "valid, instance",
    [
        (
            True,
            {
                "description": "test group with schema",
                "schema": {},
                "children": [TEST],
            },
        ),
        (
            True,
            {
                "description": "test group with schema and registry",
                "schema": {},
                "registry": {"urn:example": {}},
                "children": [TEST],
            },
        ),
        (
            True,
            {
                "description": "outer test group",
                "children": [
                    {
                        "description": "inner test group",
                        "schema": {},
                        "children": [TEST],
                    },
                ],
            },
        ),
        (
            True,
            {
                "description": "level 3",
                "children": [
                    {
                        "description": "level 2",
                        "children": [
                            {
                                "description": "level 1",
                                "schema": {},
                                "children": [TEST],
                            },
                        ],
                    },
                ],
            },
        ),
        (
            False,
            {
                "description": "test group missing schema everywhere",
                "children": [
                    {
                        "description": "a test",
                        "instance": 37,
                    },
                ],
            },
        ),
        (
            False,
            {
                "description": "duplicated schema",
                "schema": {},
                "children": [
                    {
                        "description": "inner test group",
                        "schema": {},
                        "children": [TEST],
                    },
                ],
            },
        ),
        (
            False,
            {
                "description": "duplicated registry",
                "schema": {},
                "registry": {},
                "children": [
                    {
                        "description": "inner test group",
                        "registry": {"urn:example": {}},
                        "children": [TEST],
                    },
                ],
            },
        ),
        (
            False,
            {
                "description": "nonhomogeneous test group",
                "children": [
                    TEST,
                    {
                        "description": "test group, not test",
                        "children": [ANOTHER_TEST],
                    },
                ],
            },
        ),
        (
            False,
            {
                "description": "test group with invalid registry",
                "schema": {},
                "registry": 37,
                "children": [TEST],
            },
        ),
        (
            False,
            {
                "description": "test group with empty children",
                "schema": {},
                "children": [],
            },
        ),
    ],
)
def test_group(valid, instance):
    registry = Direct.from_id("python-jsonschema").registry()
    validator = registry.for_uri("tag:bowtie.report,2023:models:group")
    assert validator.is_valid(instance) == valid, validator.validate(instance)


def test_root_schema():
    registry = Direct.from_id("python-jsonschema").registry()
    canonical_url = "tag:bowtie.report,2023:ihop"
    schema = registry.schema(canonical_url)
    assert schema["$id"] == canonical_url



from hypothesis import HealthCheck, given, settings
from hypothesis.strategies import sets
import pytest

from bowtie import HOMEPAGE, REPO
from bowtie._commands import CaseResult, SeqCase, SeqResult, TestResult
from bowtie._core import Dialect, Example, ImplementationInfo, TestCase
from bowtie._report import Report, RunMetadata
from bowtie.hypothesis import dialects, implementations, known_dialects

TestCase.__test__ = False  # frigging py.test


DIALECT_2020, DIALECT_2019 = (
    Dialect.by_alias()["2020"],
    Dialect.by_alias()["2019"],
)
FOO = ImplementationInfo(
    name="foo",
    language="blub",
    homepage=HOMEPAGE,
    issues=REPO / "issues",
    source=REPO,
    dialects=frozenset([DIALECT_2020]),
)
BAR = ImplementationInfo(
    name="bar",
    language="crust",
    homepage=HOMEPAGE,
    issues=REPO / "issues",
    source=REPO,
    dialects=frozenset([DIALECT_2020]),
)
BAZ_V1 = ImplementationInfo(
    name="baz",
    language="quux",
    version="1",
    homepage=HOMEPAGE,
    issues=REPO / "issues",
    source=REPO,
    dialects=frozenset([DIALECT_2020, DIALECT_2019]),
)
BAZ_V2 = ImplementationInfo(
    name="baz",
    language="quux",
    version="2",
    homepage=HOMEPAGE,
    issues=REPO / "issues",
    source=REPO,
    dialects=frozenset([DIALECT_2020, DIALECT_2019]),
)
FOO_RUN = RunMetadata(dialect=DIALECT_2020, implementations={"foo": FOO})
NO_FAIL_FAST = dict(did_fail_fast=False)


def test_eq():
    data = (
        FOO_RUN.serializable(),
        SeqCase(
            seq=1,
            case=TestCase(
                description="foo",
                schema={},
                tests=[Example(description="1", instance=1)],
            ),
        ).serializable(),
        SeqResult(
            seq=1,
            implementation="foo",
            expected=[None],
            result=CaseResult(results=[TestResult.VALID]),
        ).serializable(),
        SeqCase(
            seq=2,
            case=TestCase(
                description="bar",
                schema={},
                tests=[Example(description="1", instance="bar")],
            ),
        ).serializable(),
        NO_FAIL_FAST,
    )
    assert Report.from_input(data) == Report.from_input(data)


def test_eq_different_seqs():
    data = [
        FOO_RUN.serializable(),
        SeqCase(
            seq=1,
            case=TestCase(
                description="foo",
                schema={},
                tests=[Example(description="1", instance=1)],
            ),
        ).serializable(),
        SeqResult(
            seq=1,
            implementation="foo",
            expected=[None],
            result=CaseResult(results=[TestResult.VALID]),
        ).serializable(),
        SeqCase(
            seq=2,
            case=TestCase(
                description="bar",
                schema={},
                tests=[Example(description="1", instance="bar")],
            ),
        ).serializable(),
        NO_FAIL_FAST,
    ]
    reseqed = [
        FOO_RUN.serializable(),
        SeqCase(
            seq=1,
            case=TestCase(
                description="foo",
                schema={},
                tests=[Example(description="1", instance=1)],
            ),
        ).serializable(),
        SeqResult(
            seq=1,
            implementation="foo",
            expected=[None],
            result=CaseResult(results=[TestResult.VALID]),
        ).serializable(),
        SeqCase(
            seq=200,
            case=TestCase(
                description="bar",
                schema={},
                tests=[Example(description="1", instance="bar")],
            ),
        ).serializable(),
        NO_FAIL_FAST,
    ]
    assert Report.from_input(data) == Report.from_input(reseqed)


def test_eq_different_order():
    data = [
        FOO_RUN.serializable(),
        SeqCase(
            seq=1,
            case=TestCase(
                description="foo",
                schema={},
                tests=[Example(description="1", instance=1)],
            ),
        ).serializable(),
        SeqResult(
            seq=1,
            implementation="foo",
            expected=[None],
            result=CaseResult(results=[TestResult.VALID]),
        ).serializable(),
        SeqCase(
            seq=2,
            case=TestCase(
                description="bar",
                schema={},
                tests=[Example(description="1", instance="bar")],
            ),
        ).serializable(),
        NO_FAIL_FAST,
    ]
    reordered = [
        FOO_RUN.serializable(),
        SeqCase(
            seq=2,
            case=TestCase(
                description="bar",
                schema={},
                tests=[Example(description="1", instance="bar")],
            ),
        ).serializable(),
        SeqCase(
            seq=1,
            case=TestCase(
                description="foo",
                schema={},
                tests=[Example(description="1", instance=1)],
            ),
        ).serializable(),
        SeqResult(
            seq=1,
            implementation="foo",
            expected=[None],
            result=CaseResult(results=[TestResult.VALID]),
        ).serializable(),
        NO_FAIL_FAST,
    ]
    assert Report.from_input(data) == Report.from_input(reordered)


@pytest.mark.xfail(reason="We should use some other structure for results.")
def test_eq_different_seqs_different_order():
    data = [
        FOO_RUN.serializable(),
        SeqCase(
            seq=1,
            case=TestCase(
                description="foo",
                schema={},
                tests=[Example(description="1", instance=1)],
            ),
        ).serializable(),
        SeqResult(
            seq=1,
            implementation="foo",
            expected=[None],
            result=CaseResult(results=[TestResult.VALID]),
        ).serializable(),
        SeqCase(
            seq=2,
            case=TestCase(
                description="bar",
                schema={},
                tests=[Example(description="1", instance="bar")],
            ),
        ).serializable(),
        NO_FAIL_FAST,
    ]
    different_seqs = [
        FOO_RUN.serializable(),
        SeqCase(
            seq=200,
            case=TestCase(
                description="bar",
                schema={},
                tests=[Example(description="1", instance="bar")],
            ),
        ).serializable(),
        SeqCase(
            seq=100,
            case=TestCase(
                description="foo",
                schema={},
                tests=[Example(description="1", instance=1)],
            ),
        ).serializable(),
        SeqResult(
            seq=100,
            implementation="foo",
            expected=[None],
            result=CaseResult(results=[TestResult.VALID]),
        ).serializable(),
        NO_FAIL_FAST,
    ]
    assert Report.from_input(data) == Report.from_input(different_seqs)


def test_ne_different_results():
    data = [
        FOO_RUN.serializable(),
        SeqCase(
            seq=1,
            case=TestCase(
                description="foo",
                schema={},
                tests=[Example(description="1", instance=1)],
            ),
        ).serializable(),
        SeqResult(
            seq=1,
            implementation="foo",
            expected=[None],
            result=CaseResult(results=[TestResult.VALID]),
        ).serializable(),
        SeqCase(
            seq=2,
            case=TestCase(
                description="bar",
                schema={},
                tests=[Example(description="1", instance="bar")],
            ),
        ).serializable(),
        NO_FAIL_FAST,
    ]
    different_result = [
        FOO_RUN.serializable(),
        SeqCase(
            seq=1,
            case=TestCase(
                description="foo",
                schema={},
                tests=[Example(description="1", instance=1)],
            ),
        ).serializable(),
        SeqResult(
            seq=1,
            implementation="foo",
            expected=[None],
            result=CaseResult(results=[TestResult.INVALID]),
        ).serializable(),
        SeqCase(
            seq=2,
            case=TestCase(
                description="bar",
                schema={},
                tests=[Example(description="1", instance="bar")],
            ),
        ).serializable(),
        NO_FAIL_FAST,
    ]

    assert Report.from_input(data) != Report.from_input(different_result)


def test_eq_combine_versioned_reports():
    BAZ_V1_2020_RUN = RunMetadata(
        dialect=DIALECT_2020,
        implementations={"baz_v1": BAZ_V1},
    )
    BAZ_V1_2019_RUN = RunMetadata(
        dialect=DIALECT_2019,
        implementations={"baz_v1": BAZ_V1},
    )
    data_baz_v1_2020 = [
        BAZ_V1_2020_RUN.serializable(),
        SeqCase(
            seq=1,
            case=TestCase(
                description="foo",
                schema={},
                tests=[Example(description="1", instance=1)],
            ),
        ).serializable(),
        SeqResult(
            seq=1,
            implementation="baz_v1",
            expected=[None],
            result=CaseResult(results=[TestResult.VALID]),
        ).serializable(),
        NO_FAIL_FAST,
    ]
    data_baz_v1_2019 = [
        BAZ_V1_2019_RUN.serializable(),
        SeqCase(
            seq=1,
            case=TestCase(
                description="bar",
                schema={},
                tests=[Example(description="1", instance="bar")],
            ),
        ).serializable(),
        SeqResult(
            seq=1,
            implementation="baz_v1",
            expected=[None],
            result=CaseResult(results=[TestResult.VALID]),
        ).serializable(),
        NO_FAIL_FAST,
    ]

    BAZ_V2_2020_RUN = RunMetadata(
        dialect=DIALECT_2020,
        implementations={"baz_v2": BAZ_V2},
    )
    data_baz_v2_2020 = [
        BAZ_V2_2020_RUN.serializable(),
        SeqCase(
            seq=1,
            case=TestCase(
                description="foo",
                schema={},
                tests=[Example(description="1", instance=1)],
            ),
        ).serializable(),
        SeqResult(
            seq=1,
            implementation="baz_v2",
            expected=[None],
            result=CaseResult(results=[TestResult.INVALID]),
        ).serializable(),
        NO_FAIL_FAST,
    ]

    combined = [
        RunMetadata(
            dialect=DIALECT_2020,
            implementations={
                "baz_v1": BAZ_V1,
                "baz_v2": BAZ_V2,
            },
        ).serializable(),
        SeqCase(
            seq=1,
            case=TestCase(
                description="foo",
                schema={},
                tests=[Example(description="1", instance=1)],
            ),
        ).serializable(),
        SeqResult(
            seq=1,
            implementation="baz_v1",
            expected=[None],
            result=CaseResult(results=[TestResult.VALID]),
        ).serializable(),
        SeqResult(
            seq=1,
            implementation="baz_v2",
            expected=[None],
            result=CaseResult(results=[TestResult.INVALID]),
        ).serializable(),
        NO_FAIL_FAST,
    ]

    assert Report.combine_versioned_reports_for(
        [
            Report.from_input(data_baz_v1_2020),
            Report.from_input(data_baz_v1_2019),
            Report.from_input(data_baz_v2_2020),
        ],
        dialect=DIALECT_2020,
    ) == Report.from_input(combined)


@given(dialect=known_dialects)
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_ne_different_implementations(dialect):
    foo = RunMetadata(
        dialect=dialect,
        implementations={"foo": FOO},
    )
    foo_and_bar = RunMetadata(
        dialect=dialect,
        implementations={"foo": FOO, "x/baz": BAR},
    )
    data = [
        SeqCase(
            seq=1,
            case=TestCase(
                description="foo",
                schema={},
                tests=[Example(description="1", instance=1)],
            ),
        ).serializable(),
        SeqResult(
            seq=1,
            implementation="foo",
            expected=[None],
            result=CaseResult(results=[TestResult.VALID]),
        ).serializable(),
        SeqCase(
            seq=2,
            case=TestCase(
                description="bar",
                schema={},
                tests=[Example(description="1", instance="bar")],
            ),
        ).serializable(),
        NO_FAIL_FAST,
    ]

    foo_report = Report.from_input([foo.serializable(), *data])
    assert Report.from_input([foo_and_bar.serializable(), *data]) != foo_report


@given(dialect=dialects())
def test_eq_different_bowtie_version(dialect):
    one = Report.empty(dialect=dialect, bowtie_version="1970-1-1")
    two = Report.empty(dialect=dialect, bowtie_version="2000-12-31")
    assert one == two


@given(dialects=sets(dialects(), min_size=2, max_size=2))
def test_ne_different_dialect(dialects):
    one, two = dialects
    assert Report.empty(dialect=one) != Report.empty(dialect=two)


@given(dialect=dialects())
def test_eq_empty(dialect):
    assert Report.empty(dialect=dialect) == Report.empty(dialect=dialect)


@given(dialect=dialects())
def test_empty_is_empty(dialect):
    report = Report.empty(dialect=dialect)
    assert report.is_empty


@given(dialect=dialects(), implementations=implementations(min_size=0))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_empty_with_implementations_is_empty(dialect, implementations):
    report = Report.empty(dialect=dialect, implementations=implementations)
    assert report.is_empty



"""
Tests for extracting information out of GitHub URLs.
"""

import pytest

from bowtie._suite import path_and_ref_from_gh_path


@pytest.mark.parametrize(
    "path, expected",
    [
        (
            "tree/main/tests/draft2020-12",
            ("tests/draft2020-12", "main"),
        ),
        (
            "tree/main/tests/draft2020-12/",
            ("tests/draft2020-12", "main"),
        ),
        (
            "blob/main/tests/draft2020-12/contains.json",
            ("tests/draft2020-12/contains.json", "main"),
        ),
        (
            "blob/branch/with/slashes/tests/draft2020-12/contains.json",
            ("tests/draft2020-12/contains.json", "branch/with/slashes"),
        ),
        (
            "blob/CASESENSITIVEBRANCH/tests/draft2020-12/contains.json",
            ("tests/draft2020-12/contains.json", "CASESENSITIVEBRANCH"),
        ),
        # this URL isn't real, in that you get tree/ not blob/,
        # but we can easily support it anyhow in case someone manually
        # edits the URL
        (
            "blob/main/tests/draft2020-12/",
            ("tests/draft2020-12", "main"),
        ),
    ],
)
def test_path_and_ref(path, expected):
    assert path_and_ref_from_gh_path(path.split("/")) == expected



from pathlib import Path
import random
import string

from url.url import URL

from bowtie._benchmarks import Benchmark, BenchmarkGroup
from bowtie._core import Dialect


def get_benchmark():
    name = "pattern"
    benchmark_type = "keyword"
    description = (
        "A benchmark for measuring performance of the "
        "implementation for the pattern keyword."
    )
    max_string_length = 100000
    varying_parameter = "String length"

    string_length = 1000
    benchmarks = []
    while string_length <= max_string_length:
        letters = string.ascii_letters
        random_letter_string = "".join(
            random.choice(letters) for _ in range(string_length)
        )

        tests = (
            [
                dict(description="Empty String", instance=""),
                dict(
                    description="Invalid Char at First",
                    instance="1" + random_letter_string,
                ),
                dict(
                    description="Invalid Char at Middle",
                    instance=(
                        random_letter_string[: string_length // 2]
                        + "1"
                        + random_letter_string[string_length // 2 :]
                    ),
                ),
                dict(
                    description="Invalid Char at Last",
                    instance=random_letter_string + "1",
                ),
                dict(
                    description="Valid String",
                    instance=random_letter_string,
                ),
            ]
            if string_length == max_string_length
            else [
                dict(
                    description="Valid String",
                    instance=random_letter_string,
                ),
            ]
        )

        benchmarks.append(
            Benchmark.from_dict(
                name=f"String length - {string_length}",
                description=(
                    f"Validating the `pattern` keyword over string of length {string_length}."
                ),
                schema=dict(type="string", pattern="^[a-zA-Z]+$"),
                tests=tests,
            ),
        )
        string_length *= 10

    return BenchmarkGroup(
        name=name,
        benchmark_type=benchmark_type,
        description=description,
        dialects_supported=[
            Dialect.from_str("https://json-schema.org/draft/2020-12/schema"),
            Dialect.from_str("https://json-schema.org/draft/2019-09/schema"),
            Dialect.from_str("http://json-schema.org/draft-07/schema#"),
            Dialect.from_str("http://json-schema.org/draft-06/schema#"),
            Dialect.from_str("http://json-schema.org/draft-04/schema#"),
            Dialect.from_str("http://json-schema.org/draft-03/schema#"),
        ],
        benchmarks=benchmarks,
        uri=URL.parse(Path(__file__).absolute().as_uri()),
        varying_parameter=varying_parameter,
    )
